from __future__ import annotations
from ._utils import _developer_utils
from .util import find_seasonal_length
from .typing_utils import ScikitLike, NormalizerLike
from .types import (
    DynamicTesting, 
    XvarValues, 
    PositiveInt, 
    ConfInterval, 
    ModelValues, 
    DetermineBestBy, 
)
from .classes import AR
from typing import TYPE_CHECKING, Self, Optional, Any, Literal, Sequence
import pandas as pd
import numpy as np
if TYPE_CHECKING:
    from .Forecaster import Forecaster
    from .MVForecaster import MVForecaster

class SKLearnUni:
    """ Model class that supports any scikit-learn API estimator for univariate forecasting.

    Args:
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (Scikit-learn API Estimator): The imported scikit-learn API regression estimator/class (such as LinearRegressor or XGBRegressor).
        dynamic_testing (bool or int): Whether to dynamically test the model or how many steps. Ignored when test_set_actuals not specified.
        Xvars (list[str]): List of regressors to use from the passed Forecaster object.
        normalizer (NormalizerLike): Default 'minmax'. The label of the normalizer to use.
        test_set_actuals (list[float]): Optional. Test-set actuals to use for testing the model. This enables the dynamic_testing option.
        **kwargs: Passed to the scikit-learn model passed to model.
    """
    def __init__(
        self, 
        f:'Forecaster', 
        model:ScikitLike, 
        dynamic_testing:DynamicTesting = True, 
        Xvars:XvarValues = None, 
        normalizer:NormalizerLike = 'minmax', 
        test_set_actuals:Optional[list[float]]=None,
        **kwargs:Any,
    ):
        self.regr = model(**kwargs)
        self.f = f
        self.normalizer = normalizer
        self.scaler = self.f.lookup_normalizer(normalizer)()
        self.dynamic_testing = self._parse_dynamic_testing(dynamic_testing)
        self.Xvars = self._parse_Xvars(Xvars)
        self.max_lag_order = self._determine_max_lag_order(self.Xvars)
        self.test_set_actuals = test_set_actuals
        self.current_actuals = f.y.to_list()
            
    def _parse_dynamic_testing(self,dynamic_testing) -> int:
        steps = len(self.f.future_dates)
        match dynamic_testing:
            case False|None:
                return 1
            case True:
                return steps+1
            case i if i <= 0:
                raise ValueError(f'Invalid value passed to dynamic_testing: {dynamic_testing}')
            case _:
                return dynamic_testing
            
    def _determine_max_lag_order(self,Xvars):
        lag_orders = [x.lag_order for x in Xvars if isinstance(x,AR)]
        if lag_orders:
            return max(lag_orders)
        return 0
    
    def _parse_Xvars(self, Xvars):
        match Xvars:
            case 'all'|None:
                return list(self.f.current_xreg.keys())
            case _:
                return list(Xvars)

    def generate_current_X(self) -> np.ndarray:
        """ Returns the matrix of the current input dataset.
        """
        obs_to_drop = self.max_lag_order
        X = np.array([self.f.current_xreg[x].values[obs_to_drop:].copy() for x in self.Xvars]).T
        self.scaler = self.scaler.fit(X)
        return X

    def generate_future_X(self) -> np.ndarray:
        """ Returns the matrix of the future input dataset. 
        """
        X = np.array([np.array(self.f.future_xreg[x][:]) for x in self.Xvars]).T
        return X

    @_developer_utils.log_warnings
    def fit(self,X:np.ndarray,y:np.ndarray,**fit_params:Any) -> Self:
        """ Fits the estimator.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals.
            **fit_params: Passed to the .fit() method from the scikit-learn model.

        Returns:
            Self
        """
        obs_to_drop = self.max_lag_order
        X = self.scaler.transform(X)
        self.regr.fit(X,np.asarray(y)[obs_to_drop:],**fit_params)
        return self

    def predict(self,X:np.ndarray,in_sample:bool=False,**predict_params:Any) -> list[float]:
        """ Makes predictions.

        Args:
            X (np.ndarray): The input data.
            in_sample (bool): Default False. If True, returns fitted values with a one-step ahead forecast.
            **predict_params: Passed to the estimator's predict() method.
        
        Returns:
            list[float]: The predictions.
        """
        if self.max_lag_order == 0 or in_sample:
            X = self.scaler.transform(X)
            return list(self.regr.predict(X,**predict_params))
        
        if self.test_set_actuals:
            peeks = [a if (i+1)%self.dynamic_testing == 0 else np.nan for i, a in enumerate(self.test_set_actuals)]
        else:
            peeks = []
        
        series = self.current_actuals # this is used to add peeking to the models

        preds = [] # this is used to produce the real predictions
        n_steps, n_feat = X.shape
        for i in range(n_steps):
            p = self.scaler.transform(X[i,:].reshape(1,n_feat))
            pred = self.regr.predict(p,**predict_params)[0]
            preds.append(pred)
            if peeks and not np.isnan(peeks[i]):
                series.append(peeks[i])
            else:
                series.append(pred)

            if i == (n_steps-1):
                break

            for pos, x in enumerate(self.Xvars):
                if isinstance(x,AR):
                    X[i+1,pos] = series[-x.lag_order]

        return list(preds)

    def fit_predict(self,X:np.ndarray,y:np.ndarray) -> list[float]:
        """ Runs fit and predict methods, returning predictions.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals.

        Returns:
            list[float]: The predictions.
        """
        self.fit(X,y)
        return self.predict(X)
    
class SKLearnMV:
    """ Model class that supports any scikit-learn API estimator for multivariate forecasting.

    Args: 
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (Scikit-learn API Estimator): The imported scikit-learn API regression estimator/class (such as LinearRegressor or XGBRegressor).
        lags (None or int or list[int] or dict[str,int or list[int]]): The number of lags to add to the model.
            If int, that many lags added to every model.
            If a list of ints, only the lags in the list are added.
            If dict, key is a series name and value is int or list of ints that follows the behavior descrbied above, but only targeting passed series.
        dynamic_testing (bool or int): Whether to dynamically test the model or how many steps. Ignored when test_set_actuals not specified.
        Xvars (list[str]): List of regressors to use from the passed Forecaster object.
        normalizer (NormalizerLike): Default 'minmax'. The label of the normalizer to use.
        test_set_actuals (list[float]): Optional. Test-set actuals to use for testing the model. This enables the dynamic_testing option.
        **kwargs: Passed to the scikit-learn model passed to model.
    """
    def __init__(
        self,
        f:'MVForecaster', 
        model:ScikitLike, 
        lags:None|int|list[int]|dict[str,int|list[int]]=1,
        dynamic_testing:DynamicTesting = True, 
        Xvars:XvarValues = 'all', 
        normalizer:NormalizerLike = 'minmax', 
        test_set_actuals:Optional[dict[str,list[float]]]=None,
        **kwargs:Any,
    ):
        self.regr = {label:model(**kwargs) for label in f.y}
        self.f = f
        self.normalizer = normalizer
        self.lags = self._parse_lags(lags)
        self.scaler = self.f.lookup_normalizer(normalizer)()
        self.dynamic_testing = self._parse_dynamic_testing(dynamic_testing)
        self.Xvars = self._parse_Xvars(Xvars)
        self.test_set_actuals = test_set_actuals
        self.current_actuals = f.y

    def _parse_dynamic_testing(self,dynamic_testing) -> int:
        steps = len(self.f.future_dates)
        match dynamic_testing:
            case False|None:
                return 1
            case True:
                return steps+1
            case i if i <= 0:
                raise ValueError(f'Invalid value passed to dynamic_testing: {dynamic_testing}')
            case _:
                return dynamic_testing

    def _parse_lags(self,lags):
        match lags:
            case None|0:
                return 0
            case str():
                raise ValueError(f'Unrecognized value passed to lags: {lags}')
            case float():
                return int(lags)
            case _:
                return lags

    def _parse_Xvars(self,Xvars):
        match Xvars:
            case 'all':
                return list(self.f.current_xreg.keys())
            case None:
                return []
            case _:
                return list(Xvars)
            
    def _generate_X(self):
        self.predict_with_Xvars = self.Xvars[:]
        observed = np.array([self.f.current_xreg[x].values.copy() for x in self.Xvars]).T
        future = np.array([np.array(self.f.future_xreg[x][:]) for x in self.Xvars]).T

        ylen = len(self.f.y[self.f.names[0]])
        
        no_other_xvars = not observed.shape[0]
        if no_other_xvars:
            observed_future = np.array([0]*(ylen + len(self.f.future_dates))).reshape(-1,1) # column of 0s
            observed = np.array([0]*ylen).reshape(-1,1)
        else:
            observed_future = np.concatenate([observed,future],axis=0)

        if not self.lags: # handle none
            observedy = np.array([v.to_list() for _, v in self.f.y.items()]).T
            futurey = np.zeros((len(self.f.future_dates),self.f.n_series))
            if no_other_xvars:
                observed = observedy
                future = futurey
            else:
                observed = np.concatenate([observedy,observed],axis=1)
                future = np.concatenate([futurey,future],axis=1)
            
            self.current_X = observed
            self.future_X = future
            return observed, future
        elif isinstance(self.lags,int): # handle int case
            max_lag = self.lags
            lag_matrix = np.zeros((observed_future.shape[0],max_lag*self.f.n_series))

            pos = 0
            for i in range(self.f.n_series):
                for j in range(self.lags):
                    self.predict_with_Xvars.append(f'LAG_{self.f.names[i]}_{j+1}') # LAG_UTUR_1 for first lag to keep track of position
                    lag_matrix[:,pos] = (
                        [np.nan] * (j+1)
                        + self.f.y[self.f.names[i]].to_list() 
                        + [np.nan] * (lag_matrix.shape[0] - ylen - (j+1)) # pad with nas
                    )[:lag_matrix.shape[0]] 
                    pos += 1
        elif isinstance(self.lags,dict): # handle dict case
            total_lags = 0
            for _, lag_val in self.lags.items():
                local_lags = self._parse_lags(lag_val)
                if hasattr(local_lags,'__len__'):
                    total_lags += len(local_lags)
                else:
                    total_lags += local_lags
            lag_matrix = np.zeros((observed_future.shape[0],total_lags))
            pos = 0
            max_lag = 1
            
            for label, lag_val in self.lags.items():
                if hasattr(lag_val,'__len__'):
                    for i in lag_val:
                        lag_matrix[:,pos] = (
                            [np.nan] * i
                            + self.f.y[k].to_list()
                            + [np.nan]
                            * (lag_matrix.shape[0] - ylen - i)
                        )[:lag_matrix.shape[0]] 
                        self.predict_with_Xvars.append(f'LAG_{label}_{i}')
                        pos+=1
                    max_lag = max(max_lag,max(lag_val))
                else:
                    for i in range(lag_val):
                        lag_matrix[:,pos] = (
                            [np.nan] * (i+1)
                            + self.f.y[k].to_list()
                            + [np.nan]
                            * (lag_matrix.shape[0] - ylen - (i+1))
                        )[:lag_matrix.shape[0]] 
                        self.predict_with_Xvars.append(f'LAG_{label}_{i+1}')
                        pos+=1
                    max_lag = max(max_lag,lag_val)
        
        else:
            lag_matrix = np.zeros((observed_future.shape[0],len(self.lags)*self.f.n_series))
            pos = 0
            max_lag = max(self.lags)
            for i in range(self.f.n_series):
                for lag in self.lags:
                    self.predict_with_Xvars.append(f'LAG_{self.f.names[i]}_{lag}')
                    lag_matrix[:,pos] = (
                        [np.nan] * lag
                        + self.f.y[self.f.names[i]].to_list()
                        + [np.nan] * (lag_matrix.shape[0] - ylen - lag)
                    )[:lag_matrix.shape[0]]
                    pos+=1

        observed_future = np.concatenate([observed_future,lag_matrix],axis=1)

        if no_other_xvars:
            start_col = 1
        else:
            start_col = 0

        future = observed_future[observed.shape[0]:,start_col:]
        observed = observed_future[max_lag:observed.shape[0],start_col:]

        self.current_X = observed
        self.future_X = future

        return observed, future

    def generate_current_X(self) -> np.ndarray:
        """ Returns the matrix of the current input dataset.
        """
        X = self._generate_X()[0]
        self.scaler = self.scaler.fit(X)
        return X

    def generate_future_X(self) -> np.ndarray:
        """ Returns the matrix of the future input dataset. 
        """       
        return self._generate_X()[1]

    @_developer_utils.log_warnings
    def fit(self,X,y,**fit_params):
        """ Fits the estimator.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals.
            **fit_params: Passed to the .fit() method from the scikit-learn model.

        Returns:
            Self
        """
        for label, series in y.items():
            X_scaled = self.scaler.transform(X)
            self.regr[label].fit(X_scaled,series[-X.shape[0]:].copy(),**fit_params)
        return self

    def predict(self,X,in_sample:bool=False,**predict_params) -> dict[str,list[float]]:
        """ Makes predictions.

        Args:
            X (np.ndarray): The input data.
            in_sample (bool): Default False. If True, returns fitted values with a one-step ahead forecast.
            **predict_params: Passed to the estimator's predict() method.
        
        Returns:
            list[float]: The predictions.
        """       
        preds = {}
        if not self.lags or in_sample or (self.dynamic_testing == 1 and self.test_set_actuals):
            for label, regr in self.regr.items():
                X_scaled = self.scaler.transform(X)
                preds[label] = list(regr.predict(X_scaled,**predict_params))
            return preds

        series = {k: v.to_list() for k, v in self.f.y.items()}
        for i in range(X.shape[0]):
            X_scaled = self.scaler.transform(X[i,:].reshape(1,-1))
            for label, regr in self.regr.items():
                series_loc = list(self.f.y.keys()).index(label)
                pred = regr.predict(X_scaled,**predict_params)[0]
                preds.setdefault(label,[]).append(pred)
                if i < X.shape[0] - 1:
                    if (i+1) % self.dynamic_testing == 0 and self.test_set_actuals:
                        series[label].append(self.test_set_actuals[series_loc][i])
                    else:
                        series[label].append(pred)
            if i < X.shape[0] - 1:
                for x in self.predict_with_Xvars:
                    if x.startswith('LAG_'):
                        idx = self.predict_with_Xvars.index(x)
                        label = x.split('_')[1]
                        lag_no = int(x.split('_')[-1])
                        series[label][-lag_no]
                        X[i+1,idx] = series[label][-lag_no]

        return preds

    def fit_predict(self,X,y):
        """ Runs fit and predict methods, returning predictions.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals.

        Returns:
            list[float]: The predictions.
        """
        self.fit(X,y)
        return self.predict(X)
    
class VECM:
    """ Forecasts using a vector error-correction model (multivariate forecaster).

    Args: 
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (Scikit-learn API Estimator): The imported scikit-learn API regression estimator/class (such as LinearRegressor or XGBRegressor).
        lags (None or int or list[int] or dict[str,int or list[int]]): The number of lags to add to the model.
            If int, that many lags added to every model.
            If a list of ints, only the lags in the list are added.
            If dict, key is a series name and value is int or list of ints that follows the behavior descrbied above, but only targeting passed series.
        Xvars (list[str]): List of exogenous regressors to use from the passed Forecaster object. If unspecified, no regressors are used.
        test_set_actuals (list[float]): Optional. Test-set actuals to use for testing the model. This enables the dynamic_testing option.
        lags (int): The number of lags from each series to use in the model.
        coint_rank (int): Cointegration rank.
        deterministic (str): One of {"n", "co", "ci", "lo", "li"}. Default "n".
            "n" - no deterministic terms.
            "co" - constant outside the cointegration relation.
            "ci" - constant within the cointegration relation.
            "lo" - linear trend outside the cointegration relation.
            "li" - linear trend within the cointegration relation.
            Combinations of these are possible (e.g. "cili" or "colo" for linear trend with intercept). 
            When using a constant term you have to choose whether you want to restrict it to the cointegration relation 
            (i.e. "ci") or leave it unrestricted (i.e. "co"). Do not use both "ci" and "co". The same applies for "li" 
            and "lo" when using a linear term. 
        seasons (int): Default 0. Number of periods in a seasonal cycle. 0 means no seasons.
        first_season (int): Default 0. Season of the first observation.
        **kwargs: Passed to the scikit-learn model passed to model.
    """
    def __init__(
        self,
        f:'MVForecaster',
        model:Literal['auto']='auto',
        Xvars:XvarValues = None, 
        test_set_actuals:Optional[dict[str,list[float]]]=None,
        lags:int=1,
        coint_rank:int=1,
        deterministic:Literal["n", "co", "ci", "lo", "li"]="n",
        seasons:int=0,
        first_season:int=0,
        **kwargs,
    ):
        if model == 'auto':
            from statsmodels.tsa.vector_ar.vecm import VECM
        else:
            raise ValueError(f'Unrecognized value passed to model: {model}')
        
        self.f = f
        self.Xvars = self._parse_Xvars(Xvars)
        self.test_set_actuals = test_set_actuals
        self.lags = lags
        self.coint_rank = coint_rank
        self.deterministic = deterministic
        self.seasons = seasons
        self.first_season = first_season
        self.model_kwargs = kwargs
        self.init_regr = VECM

    def _parse_Xvars(self,Xvars):
        match Xvars:
            case 'all':
                return list(self.f.current_xreg.keys())
            case None:
                return []
            case _:
                return list(Xvars)
            
    def generate_current_X(self) -> np.ndarray:
        """ Returns the matrix of the current input dataset.
        """
        if self.Xvars:
            return np.array([self.f.current_xreg[x].values.copy() for x in self.Xvars]).T
        return None

    def generate_future_X(self) -> np.ndarray:
        """ Returns the matrix of the future input dataset. 
        """  
        if self.Xvars:
            return np.array([np.array(self.f.future_xreg[x][:]) for x in self.Xvars]).T
        return None

    @_developer_utils.log_warnings
    def fit(self,X:np.ndarray|None,y:np.ndarray,**fit_params):
        """ Fits the estimator.

        Args:
            X (np.ndarray): The exogenours input data. None is an accepted value.
            y (np.ndarray): The observed actuals.
            **fit_params: Passed to the .fit() method from the scikit-learn model.

        Returns:
            Self
        """
        y = np.array([v.values.copy() for v in self.f.y.values()]).T
        self.regr = (
            self.init_regr(
                y, 
                exog = X, 
                k_ar_diff=self.lags,
                coint_rank=self.coint_rank,
                deterministic=self.deterministic,
                seasons=self.seasons,
                first_season=self.first_season,
                dates=self.f.current_dates,
                **self.model_kwargs,
            )
            .fit(**fit_params)
        )
        return self

    def predict(self,X:np.ndarray,in_sample:bool=False,**predict_params):
        """ Makes predictions.

        Args:
            X (np.ndarray): The input data.
            in_sample (bool): Default False. If True, returns fitted values with a one-step ahead forecast.
            **predict_params: Passed to the estimator's predict() method.
        
        Returns:
            list[float]: The predictions.
        """    
        if not in_sample:
            preds = self.regr.predict(steps=len(self.f.future_dates), exog_fc=X, **predict_params)
            return {name:[p[i] for p in preds] for i, name in enumerate(self.f.y)}
        else:
            return {name:[p[i] for p in self.regr.fittedvalues] for i, name in enumerate(self.f.y)}

    def fit_predict(self,X,y):
        """ Runs fit and predict methods, returning predictions.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals.

        Returns:
            list[float]: The predictions.
        """
        self.fit(X,y)
        return self.predict(X)
    
class RNN:
    """ Forecasts using a recurrent neural network model from Tensorflow.

    Args: 
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (str): Default 'auto'. 'auto' is the only accepted value.
        lags (None or int or list[int] or dict[str,int or list[int]]): The number of lags to add to the model.
            If int, that many lags added to every model.
            If a list of ints, only the lags in the list are added.
            If dict, key is a series name and value is int or list of ints that follows the behavior descrbied above, but only targeting passed series.
        Xvars (list[str]): List of regressors to use from the passed Forecaster object.
        normalizer (NormalizerLike): Default 'minmax'. The label of the normalizer to use.
        test_set_actuals (list[float]): Optional. Test-set actuals to use for testing the model. This enables the dynamic_testing option.
        layers_struct (list[tuple[str,dict[str,Union[float,str]]]]): Default [('SimpleRNN',{'units':8,'activation':'tanh'})].
            Each element in the list is a tuple with two elements.
            First element of the list is the input layer (input_shape set automatically).
            First element of the tuple in the list is the type of layer ('SimpleRNN','LSTM', or 'Dense').
            Second element is a dict.
            In the dict, key is a str representing hyperparameter name: 'units','activation', etc.
            The value is the hyperparameter value.
            See here for options related to SimpleRNN: https://www.tensorflow.org/api_docs/python/tf/keras/layers/SimpleRNN.
            For LSTM: https://www.tensorflow.org/api_docs/python/tf/keras/layers/LSTM.
            For Dense: https://www.tensorflow.org/api_docs/python/tf/keras/layers/Dense.
        loss (str or tf.keras.losses.Loss): Default 'mean_absolute_error'.
            The loss function to minimize.
            See available options here: https://www.tensorflow.org/api_docs/python/tf/keras/losses.
            Be sure to choose one that is suitable for regression tasks.
        optimizer (str or tf Optimizer): Default "Adam".
            The optimizer to use when compiling the model.
            See available values here: https://www.tensorflow.org/api_docs/python/tf/keras/optimizers.
            If str, it will use the optimizer with default args.
            If type Optimizer, will use the optimizer exactly as specified.
        learning_rate (float): Default 0.001.
            The learning rate to use when compiling the model.
            Ignored if you pass your own optimizer with a learning rate.
        random_seed (int): Optional.
            Set a seed for consistent results.
            With tensorflow networks, setting seeds does not guarantee consistent results.
        scale_X (bool): Default True.
            Whether to scale the exogenous inputs with the scaler passed to the normalizer paramater.
        scale_y (bool): Default True.
            Whether to scale the endogenous inputs (lags), as well as the model output, with the scaler passed to the normalizer paramater.
            The results will automatically return unscaled.
        **kwargs: Passed to fit() and can include epochs, verbose, callbacks, validation_split, and more.
    """
    @_developer_utils.log_warnings
    def __init__(
        self, 
        f:'Forecaster', 
        model:Literal['auto']='auto', 
        test_set_actuals:Optional[list[float]]=None, 
        Xvars:XvarValues='all',
        normalizer:NormalizerLike = 'minmax', 
        layers_struct:list[tuple[dict[str,Any]]]=[("SimpleRNN", {"units": 8, "activation": "tanh"})],
        loss:str="mean_absolute_error",
        optimizer:str="Adam",
        learning_rate:ConfInterval=0.001,
        random_seed:int=None,
        scale_y:bool = True,
        **kwargs:Any,
    ):
        if model == 'auto':
            from tensorflow.keras.models import Sequential
            from tensorflow.keras.layers import Dense, LSTM, SimpleRNN
            import tensorflow.keras.optimizers
        else:
            raise ValueError(f'Unrecognized value passed to model: {model}')
        
        self.f = f
        self.normalizer = normalizer
        self.lag_scaler = self.f.lookup_normalizer(normalizer)()
        self.scaler = self.f.lookup_normalizer(normalizer)()
        self.test_set_actuals = test_set_actuals
        self.Xvars = self._parse_Xvars(Xvars)
        self.max_lag_order = self._determine_max_lag_order(self.Xvars)
        self.layers_struct = layers_struct
        self.loss = loss
        self.learning_rate = learning_rate
        self.random_seed = random_seed
        self.scale_y = scale_y
        self.fit_kwargs = kwargs

        y = self.f.y.to_list()
        self.n_timesteps = len(self.f.future_dates)
        total_periods = self.max_lag_order + self.n_timesteps
        
        idx_end = len(y)
        idx_start = idx_end - total_periods
        y_new = []

        while idx_start > 0:
            y_line = y[idx_start + self.max_lag_order:idx_start + total_periods]
            y_new.append(y_line)
            idx_start -= 1

        self.y = np.array(y_new[::-1])
        self.lag_scaler = self.lag_scaler.fit(self.f.y.values.reshape(-1,1))
        if scale_y:
            y_pieces = []
            for col in self.y.T:
                piece = self.lag_scaler.transform(col.reshape(-1,1))
                y_pieces.append(piece)
            self.y = np.concatenate(y_pieces,axis=1)

        if isinstance(optimizer, str):
            self.optimizer = eval(f"tensorflow.keras.optimizers.{optimizer}")(learning_rate=learning_rate)
        else:
            self.optimizer = optimizer

        for i, kv in enumerate(layers_struct):
            layer_label, params = kv
            match layer_label:
                case 'SimpleRNN':
                    layer = SimpleRNN
                case 'LSTM':
                    layer = LSTM
                case 'Dense':
                    layer = Dense
                case _:
                    raise ValueError(f'Unknown model layer value passed: {layer}')
                
            if i == 0:
                if layer_label in ['LSTM','SimpleRNN']:
                    params['return_sequences'] = len(layers_struct) > 1
                model = Sequential([layer(**params, input_shape = (self.n_timesteps, 1))])
            else:
                if layer_label in ('LSTM','SimpleRNN'):
                    params['return_sequences'] = not i == (len(layers_struct) - 1)
                    if params['return_sequences']:
                        params['return_sequences'] = layers_struct[i+1][0] != 'Dense'
                model.add(layer(**params))
        model.add(Dense(self.y.shape[1]))
        self.regr = model
        self.regr.compile(optimizer=self.optimizer,loss=self.loss)

    def _determine_max_lag_order(self,Xvars):
        lag_orders = [x.lag_order for x in Xvars if isinstance(x,AR)]
        if lag_orders:
            return max(lag_orders)
        return 0
    
    def _parse_Xvars(self, Xvars):
        match Xvars:
            case 'all'|None:
                return list(self.f.current_xreg.keys())
            case _:
                return list(Xvars)
            
    def _generate_X(self):
        X_lags = np.array(
            [v.to_list() + self.f.future_xreg[k][:1] for k,v in self.f.current_xreg.items() if isinstance(k,AR) and k in self.Xvars]
        ).T
        X_other = np.array(
            [v.to_list() + self.f.future_xreg[k][:1] for k,v in self.f.current_xreg.items() if not isinstance(k,AR) and k in self.Xvars]
        ).T

        X_lags_new = X_lags[self.max_lag_order:]
        X_other_new = X_other[self.max_lag_order:]
        
        # scale lags
        if len(X_lags_new) > 0:
            X_lags_new_pieces = []
            for col in X_lags_new.T:
                new_col = self.lag_scaler.transform(col.reshape(-1,1))
                X_lags_new_pieces.append(new_col)
            X_lags_new = np.concatenate(X_lags_new_pieces,axis=1)
        # scale other regressors
        if len(X_other_new) > 0:
            X_other_train = X_other_new[:-1]
            self.scaler = self.scaler.fit(X_other_train)
            X_other_new = self.scaler.transform(X_other_new)
            
        # combine
        if len(X_lags_new) > 0 and len(X_other_new) > 0:
            X = np.concatenate([X_lags_new,X_other_new],axis=1)
        elif len(X_lags_new) > 0:
            X = X_lags_new
        else:
            X = X_other_new

        current_X = X[:-1]
        future_X = X[-1:]

        return current_X, future_X

    def generate_current_X(self) -> np.ndarray:
        """ Returns the matrix of the current input dataset.
        """
        X = self._generate_X()[0]
        if self.n_timesteps > 1:
            X = X[1:-(self.n_timesteps-1)]
        else:
            X = X[1:]

        X = X.reshape(X.shape[0], X.shape[1], 1)
        return X

    def generate_future_X(self) -> np.ndarray:
        """ Returns the matrix of the future input dataset. 
        """
        X = self._generate_X()[1]
        X = X.reshape(X.shape[0], X.shape[1], 1)
        return X

    @_developer_utils.log_warnings
    def fit(self,X:np.ndarray,y:None=None,**fit_params:Any) -> Self:
        """ Fits the estimator.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals. Ignored for RNN models since the actuals are already stored in self.y, which is used for training. 
                This is just for API consistency with other models.
            **fit_params: Passed to the .fit() method from the tensorflow model.

        Returns:
            Self
        """
        if self.random_seed is not None:
            np.random.seed(self.random_seed)
        
        self.hist = self.regr.fit(X, self.y, **self.fit_kwargs, **fit_params)
        return self

    def predict(self,X:np.ndarray,in_sample:bool=False,**predict_params) -> list[float]:
        """ Makes predictions.

        Args:
            X (np.ndarray): The input data.
            in_sample (bool): Default False. If True, returns fitted values with a one-step ahead forecast.
            **predict_params: Passed to the estimator's predict() method.
        
        Returns:
            list[float]: The predictions.
        """ 
        if in_sample:
            preds = self.regr.predict(X,**predict_params)
            preds = [p[0] for p in preds[:-1]] + [p for p in preds[-1]]
        else:
            preds = self.regr.predict(X,**predict_params)
            preds = [p for p in preds[0]]
            
        if self.scale_y:
            preds = [p[0] for p in self.lag_scaler.inverse_transform(np.asarray(preds).reshape(-1,1))]
        return preds

    def fit_predict(self,X,y) -> list[float]:
        """ Runs fit and predict methods, returning predictions.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals. Ignored for RNN models since the actuals are already stored in self.y, which is used for training.

        Returns:
            list[float]: The predictions.
        """
        self.fit(X,y)
        return self.predict(X)
    
class LSTM(RNN):
    """ Forecasts using an LSTM model from Tensorflow. Inherits from RNN and simply sets the default layer to LSTM and adds lags by adding AR terms to the Forecaster object.

    Args:
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (str): Default 'auto'. 'auto' is the only accepted value.
        lags (int): The number of lags to add to the model. Default
            1. This is added to the Forecaster object as AR terms, so the model will automatically use them as input features.
        normalizer (NormalizerLike): Default 'minmax'. The label of the normalizer to use.
        lstm_layer_sizes (list[int]): Default [8]. The number of units in each LSTM layer. The number of layers is determined by the length of the list.
        dropout (list[float]): Default [0.0]. The dropout rate to use for each LSTM layer. Should be the same length as lstm_layer_sizes. If 0, no dropout is applied.
        loss (str or tf.keras.losses.Loss): Default 'mean_absolute_error'.The loss function to minimize.
            See available options here: https://www.tensorflow.org/api_docs/python/tf/keras/losses.
            Be sure to choose one that is suitable for regression tasks.
        optimizer (str or tf Optimizer): Default "Adam". The optimizer to use when compiling the model. 
            See available values here: https://www.tensorflow.org/api_docs/python/tf/keras/optimizers.  
            If str, it will use the optimizer with default args.
            If type Optimizer, will use the optimizer exactly as specified.
        learning_rate (float): Default 0.001. The learning rate to use when compiling the model. Ignored if you pass your own optimizer with a learning rate.
        random_seed (int): Optional. Set a seed for consistent results. With tensorflow networks, setting seeds does not guarantee consistent results.
        **kwargs: Passed to fit() and can include epochs, verbose, callbacks, validation_split, and more.
    """
    def __init__(
        self,
        f:'Forecaster',
        model:Literal['auto']='auto',
        lags:PositiveInt=1,
        normalizer:NormalizerLike = 'minmax', 
        lstm_layer_sizes:Sequence[int]=[8],
        dropout:Sequence[float]=[0.0],
        loss:str="mean_absolute_error",
        activation:str="tanh",
        optimizer:str="Adam",
        learning_rate:ConfInterval=0.001,
        random_seed:Optional[int]=None,
        **kwargs:Any,
    ):
        f.add_ar_terms(lags)
        lstm_kwargs = {
            'f':f,
            'model':model,
            'Xvars':[AR(i) for i in range(1,lags+1)],
            'normalizer':normalizer,
            'lstm_layer_sizes':lstm_layer_sizes,
            'dropout':dropout,
            'loss':loss,
            'activation':activation,
            'optimizer':optimizer,
            'learning_rate':learning_rate,
            'random_seed':random_seed,
        }
        new_kwargs = {
            k: v
            for k, v in lstm_kwargs.items()
            if k not in ("lstm_layer_sizes", "dropout", "activation")
        }
        new_kwargs["layers_struct"] = [
            ("LSTM",{"units": v,"activation": lstm_kwargs["activation"],"dropout": lstm_kwargs["dropout"][i]})
            for i, v in enumerate(lstm_kwargs["lstm_layer_sizes"])
        ]
        super().__init__(**new_kwargs,**kwargs)
    
class Theta:
    """ Forecasts using a Theta model from Darts.

    Args:
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (str): The Theta model to use. Default 'auto' which selects the FourTheta model from Darts. Currently, 'auto' is the only option, but more Theta variants may be added in the future.
        test_set_actuals (list[float]): Optional. Test-set actuals to use for testing the model. This enables the dynamic_testing option.
        **kwargs: Passed to the Darts Theta model specified in model.
    """
    def __init__(
        self, 
        f:'Forecaster', 
        model:Literal['auto']='auto', 
        test_set_actuals:Optional[list[float]]=None, 
        **kwargs:Any,
    ):
        if model == 'auto':
            from darts import TimeSeries
            from darts.models.forecasting.theta import FourTheta
        else:
            raise ValueError(f'Unrecognized value passed to model: {model}')
        
        self.f = f
        self.regr = FourTheta(**kwargs)
        y = pd.Series(f.y.to_list(),f.current_dates.to_list())
        self.current_actuals = TimeSeries.from_series(y)
        self.test_set_actuals = test_set_actuals

    def generate_current_X(self) -> np.ndarray:
        """ Returns the matrix of the current input dataset.
        """
        self.current_X = self.current_actuals
        return self.current_X

    def generate_future_X(self) -> np.ndarray:
        """ Returns the matrix of the future input dataset. 
        """  
        self.future_X = [0]*len(self.f.future_dates)
        return self.future_X

    @_developer_utils.log_warnings
    def fit(self,X:np.ndarray,y:None=None,**fit_params:Any) -> Self:
        """ Fits the estimator.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals. Ignored for Theta models since the actuals are already stored in self.current_actuals, which is used for training. 
                This is just for API consistency with other models.
            **fit_params: Passed to the .fit() method from the darts model.

        Returns:
            Self
        """
        self.regr.fit(X,**fit_params)
        return self

    def predict(self,X:np.ndarray,in_sample:bool=False,**predict_params) -> list[float]:
        """
        Makes predictions.
        
        Args:
            X (np.ndarray): The input data.
            in_sample (bool): Default False. If True, returns fitted values with a one-step ahead forecast.
            **predict_params: Passed to the estimator's predict() method.
        
        Returns:
            list[float]: The predictions.
        """
        if not in_sample:
            pred = self.regr.predict(len(X),**predict_params)
            return [p[0] for p in pred.values()]
        else:
            resid = [r[0] for r in self.regr.residuals(self.current_actuals).values()]
            actuals = self.current_actuals.values().flatten()[-len(resid) :]
            return [r + a for r, a in zip(resid, actuals)]
        
    def fit_predict(self,X:np.ndarray,y:None=None) -> list[float]:
        """
        Fits and predicts on the same dataset.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals. Ignored for Theta models since the actuals are already stored in self.current_actuals, which is used for training. 
                This is just for API consistency with other models.
        Returns:
            list[float]: The predictions.
        """
        self.fit(X,y)
        return self.predict(X)

class HWES:
    """ Forecasts using a Holt-Winters Exponential Smoothing model from Statsmodels.

    Args:
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (str): The HWES model to use. Default 'auto' which selects the ExponentialSmoothing model from Statsmodels. 
            Currently, 'auto' is the only option, but more HWES variants may be added in the future.
        test_set_actuals (list[float]): Optional. Test-set actuals to use for testing the model. This enables the dynamic_testing option.
        **kwargs: Passed to the Statsmodels ExponentialSmoothing model specified in model. 
    """
    def __init__(
        self,
        f:'Forecaster', 
        model:Literal['auto']='auto', 
        test_set_actuals:Optional[list[float]]=None, 
        **kwargs:Any,
    ):
        if model == 'auto':
            from statsmodels.tsa.holtwinters import ExponentialSmoothing as HWES
        else:
            raise ValueError(f'Unrecognized value passed to model: {model}')
        self.test_set_actuals = test_set_actuals
        self.f = f
        self.model_kwargs = kwargs
        self.init_regr = HWES

    def generate_current_X(self):
        """ Placeholder method to remain consistent with other models. HWES does not use an input matrix, so this method does not need to do anything. 
            It is only included for API consistency across models.
        """
        return

    def generate_future_X(self):
        """ Placeholder method to remain consistent with other models. HWES does not use an input matrix, so this method does not need to do anything. 
            It is only included for API consistency across models.
        """
        return
    
    @_developer_utils.log_warnings
    def fit(self,X:None,y:np.ndarray,optimized:bool=True,use_brute:bool=True,**fit_params:Any) -> Self:
        """ Fits the estimator.

        Args:
            X (np.ndarray): The input data. Ignored for HWES models since HWES does not use an input matrix. This is just for API consistency with other models.
            y (np.ndarray): The observed actuals.
            optimized (bool): Default True. Whether to optimize the model's smoothing level parameters. 
                If False, the parameters will be set to the values passed in **kwargs or to the default values from the statsmodels ExponentialSmoothing model if not passed in **kwargs.
            use_brute (bool): Default True. Whether to use the brute-force optimization method when optimizing the model's smoothing level parameters. 
                This is passed to the fit() method from the statsmodels ExponentialSmoothing model and is only relevant if optimized is True. 
                    If False, the model will use the default optimization method from the statsmodels ExponentialSmoothing model when optimizing the smoothing level parameters. 
            **fit_params: Passed to the .fit() method from the statsmodels model.

        Returns:
            Self
        """
        self.regr = (
            self
            .init_regr(y, dates = self.f.current_dates, freq = self.f.freq, **self.model_kwargs)
            .fit(optimized=optimized, use_brute=use_brute, **fit_params)
        )
        return self
    
    def predict(self,X:None=None,in_sample:bool=False,**predict_params) -> list[float]:
        """ Makes predictions.

        Args:
            X (np.ndarray): The input data. Ignored for HWES models since HWES does not use an input matrix. This is just for API consistency with other models.
            in_sample (bool): Default False. If True, returns fitted values with a one-step ahead forecast.
            **predict_params: Passed to the estimator's predict() method. 

        Returns:
            list[float]: The predictions.
        """
        if in_sample:
            return list(self.regr.fittedvalues)
        else:
            pred = self.regr.forecast(len(self.f.future_dates), **predict_params)
            return list(pred)
        
    def fit_predict(self,X:None,y:np.ndarray) -> list[float]:
        """ Fits and predicts on the same dataset.

        Args:
            X (np.ndarray): The input data. Ignored for HWES models since HWES does not use an input matrix. This is just for API consistency with other models.
            y (np.ndarray): The observed actuals.
        
        Returns:
            list[float]: The predictions.
        """
        self.fit(X,y)
        return self.predict(X)

class TBATS:
    """ Forecasts using a TBATS model from the tbats package.

    Args:
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (str): The TBATS model to use. Default 'auto' which selects the TBATS model from the tbats package. Currently, 'auto' is the only option, but more TBATS variants may be added in the future.
        test_set_actuals (list[float]): Optional. Test-set actuals to use for testing the model. This enables the dynamic_testing option.
        random_seed (int): Optional. Set a seed for consistent results. With TBATS models, setting seeds does not guarantee consistent results.
        **kwargs: Passed to the TBATS model specified in model.
    """
    def __init__(
        self,
        f:'Forecaster',
        model:Literal['auto']='auto',
        test_set_actuals:Optional[list[float]]=None, 
        random_seed:Optional[int]=None,
        **kwargs:Any,
    ):
        if model == 'auto':
            from tbats import TBATS
        else:
            raise ValueError(f'Unrecognized value passed to model: {model}')
        
        self.f = f
        self.regr = TBATS(show_warnings=True,**kwargs)
        self.current_actuals = f.y.to_list()
        self.test_set_actuals = test_set_actuals
        self.random_seed = random_seed

    def generate_current_X(self) -> np.ndarray:
        """ Returns the matrix of the current input dataset.
        """
        return np.asarray(self.current_actuals)

    def generate_future_X(self) -> np.ndarray:
        """ Returns the matrix of the future input dataset. For TBATS, this is just an array of zeros equal to the length of the forecast.
        """
        return np.asarray([0]*len(self.f.future_dates))

    @_developer_utils.log_warnings
    def fit(self,X:np.ndarray,y:None=None,**fit_params:Any) -> Self:
        """ Fits the estimator.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals. Ignored for TBATS models since the actuals are already stored in self.current_actuals, which is used for training. 
                This is just for API consistency with other models.
            **fit_params: Passed to the .fit() method from the tbats model.
        
        Returns:
            Self
        """
        if self.random_seed:
            np.random.seed(self.random_seed)
        self.regr.fit(X,**fit_params)
        return self

    def predict(self,X:np.ndarray,in_sample:bool=False,**predict_params) -> list[float]:
        """ Makes predictions.

        Args:
            X (np.ndarray): The input data.
            in_sample (bool): Default False. If True, returns fitted values with a one-step ahead forecast.
            **predict_params: Passed to the estimator's predict() method.
        
        Returns:
            list[float]: The predictions.
        """
        if not in_sample:
            preds = self.regr.predict(steps = len(X),**predict_params)
            return list(preds)
        else:
            return self.regr.y_hat
        
    def fit_predict(self,X:np.ndarray,y:None=None) -> list[float]:
        """ Fits and predicts on the same dataset.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals. Ignored for TBATS models since the actuals are already stored in self.current_actuals, which is used for training. 
                This is just for API consistency with other models.
        
        Returns:
            list[float]: The predictions.
        """
        self.fit(X,y)
        return self.predict(X)


class ARIMA:
    """ Forecasts using an ARIMA model from Statsmodels.

    Args:
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (str): The ARIMA model to use. Default 'auto' which selects the ARIMA model from Statsmodels. Currently, 'auto' is the only option.
        Xvars (list[str]): List of regressors to use from the passed Forecaster object.
        test_set_actuals (list[float]): Optional. Test-set actuals to use for testing the model. This enables the dynamic_testing option.
        **kwargs: Passed to the Statsmodels ARIMA model specified in model.
    """
    def __init__(
        self,
        f:'Forecaster',
        model:Literal['auto']='auto',
        Xvars:XvarValues = None, 
        test_set_actuals:Optional[list[float]]=None, 
        **kwargs:Any,
    ):
        if model == 'auto':
            from statsmodels.tsa.arima.model import ARIMA
        else:
            raise ValueError(f'Unrecognized value passed to model: {model}')
        
        self.f = f
        self.current_actuals = f.y.to_list()
        self.test_set_actuals = test_set_actuals
        self.Xvars = self._parse_Xvars(Xvars)
        self.model_kwargs = kwargs
        self.init_regr = ARIMA

    def _parse_Xvars(self, Xvars):
        match Xvars:
            case 'all':
                return [x for x in self.f.current_xreg if not isinstance(x,AR)]
            case None:
                return []
            case _:
                return list(Xvars)

    def generate_current_X(self) -> np.ndarray:
        """ Returns the matrix of the current input exogenous variables.
            If no regressors specified using the Xvars paramater, returns None.
        """
        if self.Xvars:
            return np.array([self.f.current_xreg[x].values.copy() for x in self.Xvars]).T
        return None

    def generate_future_X(self) -> np.ndarray:
        """ Returns the matrix of the future input exogenous variables. 
            If no regressors specified using the Xvars paramater, returns None.
        """
        if self.Xvars:
            return np.array([np.array(self.f.future_xreg[x][:]) for x in self.Xvars]).T
        return None

    @_developer_utils.log_warnings
    def fit(self,X:np.ndarray,y:np.ndarray,**fit_params:Any) -> Self:
        """ Fits the estimator.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals.
            **fit_params: Passed to the .fit() method from the statsmodel model.  

        Returns:
            Self
        """
        self.regr = self.init_regr(
            y.values,
            exog=X,
            dates=self.f.current_dates.to_list(),
            freq=self.f.freq,
            **self.model_kwargs
        ).fit(**fit_params)
        return self

    def predict(
        self,
        X:np.ndarray,
        in_sample:bool=False,
        dynamic:Optional[bool]=True,
        **predict_params:Any,
    ) -> list[float]:
        """ Makes predictions.

        Args:
            X (np.ndarray): The input data.
            in_sample (bool): Default False. If True, returns fitted values with a one-step ahead forecast.
            dynamic (bool): Default True. Only relevant if in_sample is True. 
                Whether to use dynamic predictions when generating in-sample fitted values. 
                If True, uses dynamic predictions, which means that when generating fitted values for the in-sample period, 
                the model uses its own previous predictions as input rather than the actuals. 
                If False, uses one-step ahead predictions, which means that when generating fitted values for the in-sample period, 
                the model always uses the actuals from the previous time step as input rather than its own predictions. 
                Using dynamic predictions can give a better sense of out-of-sample performance since it does not rely on actuals from the in-sample period, 
                but it can also lead to worse performance since any mistakes the model makes are compounded in future predictions. 
                Using one-step ahead predictions can give a better sense of in-sample fit since it always uses the actuals from the in-sample period, 
                but it can also lead to overly optimistic performance since it relies on actuals from the in-sample period that would not be available in an out-of-sample forecasting scenario.
            **predict_params: Passed to the estimator's predict() method.
        """
        if not in_sample:
            preds = self.regr.predict(
                exog=X,
                start=len(self.current_actuals),
                end=len(self.current_actuals) + len(self.f.future_dates) - 1, 
                dynamic=dynamic,
                **predict_params,
            )
            return list(preds)
        else:
            return list(self.regr.fittedvalues)
        
    def fit_predict(self,X:np.ndarray,y:np.ndarray) -> list[float]:
        """ Fits and predicts on the same dataset.

        Args:
            X (np.ndarray): The input data.
            y (np.ndarray): The observed actuals.   

        Returns:
            list[float]: The predictions.
        """
        self.fit(X,y)
        return self.predict(X)

class Prophet:
    """ Forecasts using a Prophet model from the prophet package.

    Args:
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (str): The Prophet model to use. Default 'auto' which selects the Prophet model from the prophet package. Currently, 'auto' is the only option.
        Xvars (list[str]): List of regressors to use from the passed Forecaster object. These are added as extra regressors in the Prophet model. 
            If 'all', will use all available regressors in the Forecaster object. Default None, which uses no regressors.
        test_set_actuals (list[float]): Optional. Test-set actuals to use for testing the model. This enables the dynamic_testing option.
        cap (float): Optional. The capacity parameter to use for the Prophet model if you want to fit a logistic growth model. If not passed, the model will fit a linear growth model.
        floor (float): Optional. The floor parameter to use for the Prophet model if you want to fit a logistic growth model. If not passed, the model will fit a linear growth model.
        callback_func (callable): Optional. A function that takes the initialized but unfitted Prophet model as input and performs some operations on it, 
            such as adding holidays or changing hyperparameters, before it is fitted. 
            This allows you to customize the Prophet model in ways that are not currently supported by the parameters of this class. 
            If not passed, no operations will be performed on the initialized Prophet model before fitting.
        **kwargs: Passed to the Prophet model specified in model. Note that if you want to use the 
            dynamic_testing option with Prophet, you must pass the parameter 'interval_width' in **kwargs with a value less than 1 (e.g. 0.8) 
            to ensure that the prediction intervals are narrow enough to be useful for testing.
    """
    def __init__(
        self,
        f:'Forecaster',
        model:Literal['auto']='auto',
        Xvars:XvarValues = None, 
        test_set_actuals:Optional[list[float]]=None, 
        cap:Optional[float]=None,
        floor:Optional[float]=None,
        callback_func:callable=None,
        **kwargs:Any,
    ):
        if model == 'auto':
            from prophet import Prophet
        else:
            raise ValueError(f'Unrecognized value passed to model: {model}')
        
        self.f = f
        self.current_actuals = f.y.to_list()
        self.test_set_actuals = test_set_actuals
        self.Xvars = self._parse_Xvars(Xvars)
        self.regr = Prophet(**kwargs)
        self.cap = cap
        self.floor = floor
        self.callback_func = callback_func

    def _parse_Xvars(self, Xvars):
        match Xvars:
            case 'all':
                return [x for x in self.f.current_xreg if not isinstance(x,AR)]
            case None:
                return []
            case _:
                return list(Xvars)
            
    def generate_current_X(self) -> pd.DataFrame:
        """ Returns the DataFrame of the current input dataset. For Prophet, this includes a 'ds' column for dates, a 'y' column for the actuals, and columns for any specified regressors. 
            If cap and/or floor are specified, these are also included as columns.
        """
        df = pd.DataFrame({'ds':self.f.current_dates.to_list(),'y':self.current_actuals})
        for x in self.Xvars:
            df[x] = self.f.current_xreg[x][:]
        if self.cap:
            df['cap'] = self.cap
        if self.floor:
            df['floor'] = self.floor
        
        return df

    def generate_future_X(self) -> pd.DataFrame:
        """ Returns the DataFrame of the future input dataset. For Prophet, this includes a 'ds' column for dates and columns for any specified regressors. 
            If cap and/or floor are specified, these are also included as columns.
        """
        df = pd.DataFrame({'ds':self.f.future_dates.to_list()})
        for x in self.Xvars:
            df[x] = self.f.future_xreg[x][:]
        if self.cap:
            df['cap'] = self.cap
        if self.floor:
            df['floor'] = self.floor
        
        return df
    
    @_developer_utils.log_warnings
    def fit(self,X:pd.DataFrame,y:None=None,**fit_params:Any) -> Self:
        """ Fits the estimator.
        
        Args:
            X (pd.DataFrame): The input data.
                y (pd.DataFrame): The observed actuals. Ignored for Prophet models since the actuals are already stored in self.current_actuals, which is used for training.
                This is just for API consistency with other models.
            **fit_params: Passed to the .fit() method from the scikit-learn model.
        
        Returns:
            Self
        """
        if callable(self.callback_func):
            self.callback_func(self.regr)

        self.regr = self.regr.fit(X,**fit_params)
        return self
    
    def predict(self,X:pd.DataFrame,in_sample:None=None,**predict_params:Any) -> list[float]:
        """ Makes predictions.

        Args:
            X (pd.DataFrame): The input data.
            in_sample (bool): Ignored for Prophet models since Prophet does not have a built-in method for generating in-sample fitted values with a one-step ahead forecast. 
                This is just for API consistency with other models.
            **predict_params: Passed to the estimator's predict() method.
        Returns:
            list[float]: The predictions.
        """
        preds = self.regr.predict(X,**predict_params)
        return preds['yhat'].to_list()
    
    def fit_predict(self,X:pd.DataFrame,y:None=None) -> list[float]:
        """ Fits and predicts on the same dataset.
        
        Args:
            X (pd.DataFrame): The input data.
            y (pd.DataFrame): The observed actuals. Ignored for Prophet models since the actuals are already stored in self.current_actuals, which is used for training. 
                This is just for API consistency with other models.    
        
        Returns:
            list[float]: The predictions.
        """
        self.fit(X,y)
        return self.predict(X)

class Naive:
    """ Forecasts using a Naive model. This model simply uses the last observed value as the forecast for all future periods. 
    If seasonal, it uses the value from the same period in the previous season as the forecast.

    Args:
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (str): The Naive model to use. Must be 'auto'.
        test_set_actuals (list[float]): Not used
        seasonal (bool): Whether to use a seasonal naive model. 
            If False, the forecast for all future periods is the last observed value. 
            If True, the forecast for each future period is the value from the same period in the previous season.
        m (int or 'auto'): The seasonal period to use if seasonal is True. 
            If 'auto', the seasonal period is determined based on the frequency of the data. For example, if the frequency is monthly, the seasonal period will be 12. 
            If the frequency is quarterly, the seasonal period will be 4. If the frequency is daily, the seasonal period will be 7. 
            If the frequency is yearly, the seasonal period will be 1 (which means the seasonal naive model will be the same as the non-seasonal naive model). 
            You can also specify an integer value for the seasonal period if you want to use a different seasonal period than the one determined by the frequency.
    """
    def __init__(
        self,
        f:'Forecaster',
        model:Literal['auto']='auto',
        test_set_actuals:Optional[list[float]]=None,
        seasonal:bool=False,
        m:int|Literal['auto']='auto',
    ):
        if model != 'auto':
            raise ValueError(f'Unrecognized value passed to model: {model}')
        
        self.f = f
        self.test_set_actuals = test_set_actuals
        if seasonal:
            self.m = find_seasonal_length(m,f.freq)
        else:
            self.m = 1

    def generate_current_X(self) -> None:
        """ Placeholder method to remain consistent with other models. Naive does not use an input matrix, so this method does not need to do anything. 
            It is only included for API consistency across models.
        """
        return
    
    def generate_future_X(self) -> None:
        """ Placeholder method to remain consistent with other models. Naive does not use an input matrix, so this method does not need to do anything. 
            It is only included for API consistency across models.
        """
        return
    
    def fit(self,X:None=None,y:None=None) -> Self:
        """ Fits the estimator. For the Naive model, there is no fitting process since the model simply uses the last observed value (or the value from the same period in the previous season) as the forecast. 
            This method is included for API consistency with other models, but it does not need to do anything for the Naive model.
        
        Args:
            X (None): Ignored for the Naive model since it does not use an input matrix. This is just for API consistency with other models.
            y (None): Ignored for the Naive model since it does not use an input matrix. This is just for API consistency with other models.
        
        Returns:
            Self
        """
        return self
    
    def predict(self,X:None=None,in_sample:bool=False) -> list[float]:
        """ Makes predictions. For the Naive model, the predictions are generated by taking the last observed value (or the value from the same period in the previous season) and using it as the forecast for all future periods.
        
        Args:
            X (None): Ignored for the Naive model since it does not use an input matrix. This is just for API consistency with other models.
            in_sample (bool): Default False. If True, returns fitted values with a one-step ahead forecast. For the Naive model, the in-sample predictions are generated by taking the last observed value (or the value from the same period in the previous season) and using it as the fitted value for all periods in the in-s

        Returns:            
            list[float]: The predictions.
        """
        if in_sample:
            return self.f.y.shift(self.m).dropna().to_list()
        else:
            return (self.f.y.to_list()[-self.m:] * int(np.ceil(len(self.f.future_dates)/self.m)))[:len(self.f.future_dates)]
        
    def fit_predict(self,X:None=None,y:None=None) -> list[float]:
        """ Fits and predicts on the same dataset. For the Naive model, there is no fitting process since the model simply uses the last observed value (or the value from the same period in the previous season) as the forecast. 
            This method is included for API consistency with other models, but it does not need to do anything for the Naive model other than call the predict() method since there is no fitting process.

        Args:
            X (None): Ignored for the Naive model since it does not use an input matrix. This is just for API consistency with other models.
            y (None): Ignored for the Naive model since it does not use an input matrix. This is just for API consistency with other models.

        Returns:
            list[float]: The predictions.
        """
        self.fit(X,y)
        return self.predict(X)


class Combo:
    """ Forecasts using a combination of other evaluated forecasts from multiple models. The forecasts are combined with either simple or weighted averaging.

    
    Args:
        f (Forecaster): The Forecaster object storing the actual series and associated dates.
        model (str): The Combo model to use. Must be 'auto'.
        test_set_actuals (list[float]): Optional. Test-set actuals to use for testing the model. This enables the dynamic_testing option.
        how (str): The method to use when combining forecasts. Default 'simple' which uses simple averaging. 'weighted' uses weighted averaging where the weights are determined by the relative performance of the models on some metric.
        models (str or list[str]): The models to include in the combination. 
            Default 'all' which includes all models in the Forecaster's history except the most recent one (which is assumed to be the Combo model itself). 
            You can also specify a list of model names to include, or use the syntax 'top_n' to specify the top n models based on the metric specified in determine_best_by.
        determine_best_by (str): The metric to use when determining the best models for the 'top_n' syntax in the models argument. 
            Default 'ValidationMetricValue' which uses the validation metric value stored in the Forecaster's history for each model. 
            This is the most common use-case, but you can also specify other metrics that are stored in the history such as 'TestSetRMSE' or 'InSampleRMSE'.
        weights (list[float]): Optional. If how is 'weighted', you can optionally provide your own weights for each model instead of using the relative performance.               
        replace_negative_weights (bool|float): Whether to replace negative-scoring metrics with some positive (or 0) value to avoid situations where predictions might become nonsensical.
            This will be ignored in situations where lower scores are better (R2 is the main use-case).
            Change this to False to turn it off. 0 is an acceptable replacement value.
        exclude_models_with_no_fvs (bool): Whether to exclude models that have no fitted values stored in the history when generating the combined forecast. 
            This is relevant because if a model has no fitted values, it cannot generate in-sample predictions, 
            which means it can only contribute to the future forecast and not the in-sample fitted values. 
            This can lead to situations where the combined forecast is essentially just the forecast from that one model, which may not be desirable. Default True.
    """
    def __init__(
        self,
        f:'Forecaster',
        model:Literal['auto']='auto', 
        test_set_actuals:Optional[list[float]]=None, 
        how:Literal['simple','weighted']="simple",
        models:ModelValues="all",
        determine_best_by:DetermineBestBy="ValidationMetricValue",
        weights:Optional[Sequence[float|int]]=None,
        replace_negative_weights:bool|float=.001,
        exclude_models_with_no_fvs:bool = True,
    ):
        if model != 'auto':
            raise ValueError(f'Unrecognized value passed to model: {model}')
        
        self.f = f
        self.test_set_actuals = test_set_actuals
        self.how = how
        self.models = self._parse_models(models=models, determine_best_by=determine_best_by)
        self.weights = weights
        self.replace_negative_weights = replace_negative_weights
        if how == 'weighted' and not weights:
            self.metrics = [f.history[m][determine_best_by] for m in self.models]
            self.lower_is_better = self.metrics[0].store.lower_is_better
        elif how == 'weighted' and weights:
            if len(self.weights) != len(self.models):
                raise ValueError('When how is weighted and weights are provided, the number of provided weights must match the number of provided models')
        else:
            self.metrics = None
        self.exclude_models_with_no_fvs = exclude_models_with_no_fvs

    def _validate_how(self,how):
        if how not in ('simple','weighted'):
            raise ValueError(f'Argument passed to how not recognized: {how}')
    
    def _parse_models(self, models:ModelValues, determine_best_by:DetermineBestBy):
        all_models = [m for m in self.f.history][:-1] # exclude last model because it's the one just created (combo)
        match models:
            case 'all':
                return all_models
            case str():
                if models.startswith('top_'):
                    top_n = int(models.split('_')[-1])
                    return self.f._parse_models(models=all_models, determine_best_by=determine_best_by)[:top_n]
                else:
                    return [models]

        return models
    
    def generate_current_X(self):
        """ Generates the matrix of the current input dataset by extracting the fitted values for each model in the combination from the Forecaster's history.
        """
        lengths = []
        all_fvs = []
        for m in self.models:
            fvs = self.f.history[m].get('FittedVals',[])
            if not fvs and self.exclude_models_with_no_fvs:
                continue
            else:   
                lengths.append(len(fvs))
            all_fvs.append(fvs)
        
        min_length = min(lengths)
        return np.array([fv[-min_length:] for fv in all_fvs]).T

    def generate_future_X(self):
        """ Generates the matrix of the future input dataset by extracting either the test set predictions (if test_set_actuals were provided) 
        or the forecasts for each model in the combination from the Forecaster's history.
        """
        if self.test_set_actuals:
            return np.array([self.f.history[m]['TestSetPredictions'] for m in self.models]).T
        else:
            return np.array([self.f.history[m]['Forecast'] for m in self.models]).T
    
    @_developer_utils.log_warnings
    def fit(self,X:None=None,y:None=None,**fit_params:None) -> Self:
        """ Fits the estimator. For the Combo model, this means specifying the weights to use for combining the forecasts from the specified models based on the method specified in how.

        Args:
            X (None): Ignored for the Combo model since it does not use an input matrix for fitting. This is just for API consistency with other models.
            y (None): Ignored for the Combo model since it does not use an input matrix for fitting. This is just for API consistency with other models.
            **fit_params: Ignored for the Combo model since there is no fitting process. This is just for API consistency with other models.

        Returns:
            Self
        """
        match self.how:
            case 'simple':
                self.weights = [1/len(self.models) for _ in self.models]
            case _:
                if not self.weights:
                    weights = [m.score/sum([m.score for m in self.metrics]) for m in self.metrics]
                    if self.lower_is_better:
                        weights.reverse()
                    elif self.replace_negative_weights is not False:
                        weights = [self.replace_negative_weights if i < 0 else i for i in weights]
                    self.weights = weights
                else:
                    self.weights = [w/sum(self.weights) for w in self.weights]

        return self
    
    def predict(self,X:None=None,in_sample:bool=False,**predict_params:None) -> list[float]:
        """ Makes predictions by combining the forecasts from the specified models using either simple or weighted averaging.
        
        Args:
            X (np.ndarray): The input data. This is the matrix of forecasts from the specified models for the future periods, or the matrix of fitted values from the specified models for the in-sample period.
            in_sample (bool): Whether the predictions being generated are for the in-sample period (i.e. fitted values) or for the future forecast period. 
                This is just for API consistency with other models since the Combo model generates predictions for both the in-sample period and the 
                future forecast period in the same way by combining the forecasts from the specified models using either simple or weighted averaging.
            **predict_params: Ignored for the Combo model since there is no fitting process. This is just for API consistency with other models.
        
        Returns:
            list[float]: The combined predictions.
        """
        return list(np.sum(X * self.weights,axis=1))

    def fit_predict(self,X:None=None,y:None=None) -> list[float]:
        """ Fits and predicts on the same dataset.

        Args:
            X (None): Ignored for the Combo model since it does not use an input matrix for fitting. This is just for API consistency with other models.
            y (None): Ignored for the Combo model since it does not use an input matrix for fitting. This is just for API consistency with other models.
            **fit_params: Ignored for the Combo model since there is no fitting process. This is just for API consistency with other models.

        Returns:
            list[float]: The combined predictions.
        """
        self.fit(X,y)
        return self.predict(X)



from __future__ import annotations
from ._utils import _developer_utils, boxcox_tr, boxcox_re
from .types import ConfInterval, AvailableModel, DefaultMetric, TryTransformations, NonNegativeInt, DatetimeLike, FFillOption, PositiveInt
from .Metrics import Metrics as metrics
from .classes import EvaluatedMetric, AR
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
import seaborn as sns
import numpy as np
import pandas as pd
import warnings
import copy
from statsmodels.tsa.tsatools import freq_to_period
from typing import Sequence, Optional, Unpack, Any, Literal, TYPE_CHECKING
if TYPE_CHECKING:
    from statsmodels.tsa.vector_ar.var_model import LagOrderResults
    from statsmodels.tsa.vector_ar.vecm import CointRankResults
    from .Forecaster import Forecaster
    from .MVForecaster import MVForecaster
    from .Pipeline import Transformer, Reverter, Pipeline
    from tf.keras.callbacks import History

def plot_rnn_loss(history:'History', title:str = '', ax:Axes=None, figsize:tuple[int,int]=(12,6), show:bool = False) -> Axes:
    """ Plots the loss from a tensorflow model.

    Args:
        history (History): The model history to plot.
        title (str): Default empty. The title to apply to the plot.
        ax (Axes): The existing axis to attach the plot to.
        figsize (tuple[int,int]): Default (12,6). The size of the resulting figure.
        show (bool): Default False. Whether to show the plot.

    Returns:
        Axes: The firgure's axis.

    >>> f.set_estimator('lstm')
    >>> f.fit()
    >>> plot_rnn_loss(f.call_estimator.regr)
    """
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)
        
    ax.plot(history.history["loss"], label="train_loss")
    
    if "val_loss" in history.history.keys():
        ax.plot(history.history["val_loss"], label="val_loss")
    
    ax.set_title(title)
    ax.set_xlabel("epoch")
    ax.legend(loc="upper right")
    
    if show:
        plt.show()

    return ax


def find_seasonal_length(m:int|Literal['auto']='auto',freq:Optional[str]=None) -> int:
    """ Finds the seasonal length of a given freqency.

    Args:
        m ('auto' or int): The existing seasonal value. If you pass an int here, that same int is returned.
        freq (str): A pandas frequency value. If m is 'auto', this will be used to find the seasonal length.
    
    Returns:
        int: The inferred seasonal length.
    """
    if m == 'auto':
        if freq is not None:
            if freq.startswith('M'):
                return 12
            elif freq.startswith('Q'):
                return 4
            elif freq.startswith('H'):
                return 24
            else:
                try:
                    return freq_to_period(freq)
                except:
                    return 1
        else:
            return 1
    return m

def plot_reduction_errors(f:Forecaster, ax:Axes = None, figsize:tuple[int,int]=(12,6)) -> Axes:
    """ Plots the resulting error/accuracy of a Forecaster object where `reduce_Xvars()` method has been called
    with method = 'pfi' or method = 'shap'.
    
    Args:
        f (Forecaster): An object that has called the `reduce_Xvars()` method with method = 'pfi'.
        ax (Axis): Optional. The existing axis to write the resulting figure to.
        figsize (tuple): Default (12,6). The size of the resulting figure. Ignored when ax is not None.
        
    Returns:
        (Axis) The figure's axis.

    >>> f.reduce_Xvars()
    >>> plot_reduction_errors(f)
    >>> plt.show()
    """
    dropped = f.pfi_dropped_vars
    errors = f.pfi_error_values
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)
    sns.lineplot(
        x=np.arange(0, len(dropped) + 1, 1), y=errors,
    )
    plt.xlabel("dropped Xvars")
    plt.ylabel("error")
    return ax

def backtest_metrics(
    backtest_results: list[dict[str,pd.DataFrame]],
    models:Optional[list[str]] = None,
    mets:list[str|callable] = ['rmse'],
    mase:bool = False, 
    msis:bool = False,
    msis_alpha:ConfInterval = 0.05,
    m:int = 1, 
    names:Optional[list[str]] = None,
) -> pd.DataFrame:
    """ Ingests the results output from `Pipeline.backtest()` and converts results to metrics.
    
    Args:
        backtest_results (list): The output returned from `Pipeline.backtest()` or `MVPipeline.backtest()`.
        models (collection): The names of the models to display metrics for. Default displays all models.
        mets (list[str or callable]): Default ['rmse']. A list of metrics to calculate.
            If the element is str type, must be taken from the `util.metrics` class 
            where the only two accepted arguments are `a` and `f`.
            If the element in the list is callable, must be a function that only accepts two arguments (first actuals second forecast) 
            and returns a float.
        mase (bool): Default False.
            Whether to also calculate mase. Must specify seasonality in m.
        msis (bool): Default False.
            Whether to also calculate msis. Must specify seasonality in m. This will fail if confidence intervals were not evaluated.
        msis_alpha (float): Default 0.05. The level that confidence intervals were evaluated at. Ignored if msis is False.
        m (int): Default 1. The number of steps that count one seasonal cycle. Ignored if both of msis and mase is False.
        names (list): Optional. The names to assign each passed series. Ignored if there is only one passed series.
    
    Returns:
        (DataFrame): The metrics dataframe that gives info about each backtested series, model, and selected metric.

    >>> f1, f2, f3 = pipeline.fit_predict(f1,f2,f3)
    >>> backtest_results = pipeline.backtest(f1,f2,f3,n_iter=2,jump_back=12)
    >>> backtest_mets = backtest_metrics(
    >>>     backtest_results,
    >>>     mets = ['rmse','smape','r2','mae'],
    >>>     names=['UTUR','UNRATE','SAHMREALTIME'],
    >>>     mase = True,
    >>>     msis = True,
    >>>     m = 12,
    >>> )
    """
    m_ = m
    labels = names if names is not None else [f'Series{i}' for i in range(len(backtest_results))]
    res_dict = {k:None for k in labels}
    mets_str = [m if isinstance(m,str) else m.__name__ for m in mets]
    actuals_obs = ['Actuals','Obs']
    if models is None:
        models = [m for m in backtest_results[0].keys() if m not in actuals_obs]
    elif isinstance(models,str):
        models = [models]
    for h, s in enumerate(backtest_results):
        for i, m in enumerate(actuals_obs + models):
            if i == 0:
                a_df = s[m][[c for c in s[m] if c.endswith('Vals')]]
                results = pd.DataFrame(
                    columns = [f'Iter{i}' for i in range(a_df.shape[1])],
                    index = pd.MultiIndex.from_product(
                            [
                                models[:],
                                mets_str + ([] if not mase else ['mase']) + ([] if not msis else ['msis']),
                            ],
                            names = ['Model','Metric']
                    ),
                )
                a_df.columns = results.columns.to_list()
            elif i == 1:
                ob_df = s[m]
            else:
                f_df = s[m][[c for c in s[m] if c.endswith('Fcst')]]
                f_df.columns = results.columns.to_list()
                for met in mets_str:
                    for c in results:
                        results.loc[(m,met),c] = (
                            getattr(metrics,met)(
                                a = a_df[c],
                                f = f_df[c],
                            ) if isinstance(met,str) 
                            else met(
                                a_df[c],
                                f_df[c],
                            )
                        )
                if mase:
                    for c in results:
                        results.loc[(m,'mase'),c] = getattr(metrics,'mase')(
                            a = a_df[c],
                            f = f_df[c],
                            obs = ob_df[c].dropna(),
                            m = m_,
                        )
                if msis:
                    up_df = s[m][[c for c in s[m] if c.endswith('Upper')]]
                    lo_df = s[m][[c for c in s[m] if c.endswith('Lower')]]
                    up_df.columns = f_df.columns.to_list()
                    lo_df.columns = f_df.columns.to_list()
                    for c in results:
                        results.loc[(m,'msis'),c] = getattr(metrics,'msis')(
                            a = a_df[c],
                            uf = up_df[c],
                            lf = lo_df[c],
                            obs = ob_df[c].dropna(),
                            m = m_,
                            alpha = msis_alpha,
                        )
        res_dict[labels[h]] = results

    if len(res_dict) == 1:
        results_df = res_dict[labels[0]]
    else:
        res_list = []
        for k, df in res_dict.items():
            df['Series'] = k
            df = df.reset_index().set_index(['Series','Model','Metric'])
            res_list.append(df)
        results_df = pd.concat(res_list)

    results_df['Average'] = results_df.mean(axis=1)
    return results_df

def break_mv_forecaster(mvf:MVForecaster,drop_all_Xvars:bool = True) -> Unpack['Forecaster']:
    """ Breaks apart an MVForecaster object and returns as many Foreaster objects as series loaded into the object.

    Args:
        mvf (MVForecaster): The object to break apart.
        drop_all_Xvars (bool): Default True. Whether to drop all Xvars during the conversion.
            It's a good idea to leave this True because length mismatches can cause future univariate models
            to error out.

    Returns:
        (tuple[Forecaster]): A sequence of at least two Forecaster objects
    
    >>> from scalecast.MVForecaster import MVForecaster
    >>> from scalecast.util import break_mv_forecaster
    >>> 
    >>> f1, f2 = break_mv_forecaster(mvf)
    """
    from .Forecaster import Forecaster
    def convert_mv_hist(f, mvhist: dict, series_name: str):
        hist = {}
        for k, v in mvhist.items():
            hist[k] = {}
            for k2, v2 in v.items():
                if k2 in (""):
                    continue
                elif not isinstance(v2, dict) or k2 == "HyperParams":
                    hist[k][k2] = v2
                elif isinstance(v2, dict):
                    try:
                        hist[k][k2] = v2[series_name]
                    except IndexError:
                        hist[k][k2] = []
        return hist

    mvf1 = copy.deepcopy(mvf)

    set_len = (
        len(mvf1.y[mvf1.names[0]]) 
        if not mvf1.current_xreg 
        else len(list(mvf1.current_xreg.values())[0])
    )
    to_return = []
    for s in mvf1.names:
        f = Forecaster(
            y=mvf1.y[s].values[-set_len:],
            current_dates=mvf1.current_dates.values[-set_len:],
            future_dates=len(mvf1.future_dates),
            test_length=mvf1.test_length,
            cis = mvf1.cis,
        )
        f.current_xreg = {k:v.copy() for k,v in mvf1.current_xreg.items()}
        f.future_xreg = {k:v.copy() for k,v in mvf1.future_xreg.items()}
        f.set_validation_length(mvf1.validation_length)
        f.set_cilevel(mvf1.cilevel)
        f.history = convert_mv_hist(
            f=f, 
            mvhist=mvf1.history, 
            series_name=s,
        )
        if drop_all_Xvars:
            f.drop_all_Xvars()

        to_return.append(f)

    return tuple(to_return)

def find_optimal_lag_order(mvf:'MVForecaster',train_only:bool=False,**kwargs:Any) -> 'LagOrderResults':
    """ Returns the otpimal lag order for a mutlivariate process using the statsmodels function:
    https://www.statsmodels.org/dev/generated/statsmodels.tsa.vector_ar.var_model.VAR.select_order.html.
    The exogenous regressors are set based on Xvars loaded in the `MVForecaster` object.

    Args:
        mvf (MVForecaster): The MVForecaster object with series loaded to find the optimal order for
        train_only (bool): Default False. Whether to use the training data only in the test.
        **kwargs: Passed to the referenced statsmodels function

    Returns:
        (LagOrderResults): Lag selections.
    
    >>> from scalecast.Forecaster import Forecaster
    >>> from scalecast.MVForecaster import MVForecaster
    >>> from scalecast.util import find_optimal_lag_order
    >>> import pandas_datareader as pdr
    >>>
    >>> s1 = pdr.get_data_fred('UTUR',start='2000-01-01',end='2022-01-01')
    >>> s2 = pdr.get_data_fred('UNRATE',start='2000-01-01',end='2022-01-01')
    >>>
    >>> f1 = Forecaster(y=s1['UTUR'],current_dates=s1.index)
    >>> f2 = Forecaster(y=s2['UNRATE'],current_dates=s2.index)
    >>>
    >>> mvf = MVForecaster(f1,f2,names=['UTUR','UNRATE'])
    >>> lag_order_res = find_optimal_lag_order(mvf,train_only=True)
    >>> lag_order_aic = lag_order_res.aic # picks the best lag order according to aic
    """
    from statsmodels.tsa.vector_ar.var_model import VAR
    data = np.array([v.astype(float).values.copy() for k, v in mvf.y.items()]).T

    if mvf.current_xreg:
        exog = np.array(v.values.copy() for k, v in mvf.current_xreg.items()).T
    else:
        exog = None

    if train_only:
        data = data[:-mvf.test_length]
        if exog is not None:
            exog = exog[:-mvf.test_length]

    model = VAR(data,exog=exog)

    return model.select_order(**kwargs)

def infer_apply_Xvar_selection(infer_from:'Forecaster'|'MVForecaster',apply_to:'Forecaster'|'MVForecaster',return_copy:bool=False) -> 'Forecaster'|'MVForecaster':
    """ Attempts to infer what Xvars have been added to one Forecaster object and applies the guess to another Forecaster object.
    If using default fourier seasonal terms, linear or log trend terms, and autoregressive terms only, with defaults, 
    this will guess all variables successfully. Other variables (such as through `Forecaster.add_Xvars_df()`) will not be added. 
    Any variables that cannot be inferred will be raised in a warning.

    Args:
        infer_from (Forecaster or MVForecaster): The object to infer the Xvars from.
        apply_to (Forecaster or MVForecaster): The object to apply the guess to.
        return_copy (bool): Default False. Whether to create a copy of the object passed to `apply_to`.
            Default will add Xvars to the instance passed to `apply_to`.

    Returns:
        (Forecaster or MVForecaster): The object with the inferred variables added to it. Same object as passed to apply_to.

    >>> f2 = infer_apply_Xvar_selection(infer_from=f1,apply_to=f2)
    """
    if return_copy:
        apply_to = copy.deepcopy(apply_to)
    
    not_guessed = []
    for k in infer_from.current_xreg.keys():
        if isinstance(k,AR):
            apply_to.add_ar_terms([k.lag_order])
        elif k.endswith('sin'):
            apply_to.add_seasonal_regressors(k.split('sin')[0],sincos=True,raw=False)
        elif k.endswith('cos'):
            continue
        elif k in ('t', 'lnt','t^2'):
            apply_to.add_time_trend()
            if k == 'lnt':
                apply_to.add_logged_terms('t',drop=True)
            elif k == 't^2':
                apply_to.add_poly_terms('t')
                apply_to.drop_Xvars('t')
        else:
            not_guessed.append(k)

    if len(not_guessed):
        warnings.warn(
            f'The inference was unable to guess the following variables: {not_guessed}.'
            ' All others have been added to the Forecaster object passed to the `apply_to` argument.',
            category=Warning,
        )

    return apply_to

def find_optimal_coint_rank(mvf:'MVForecaster',det_order:int,k_ar_diff:int,train_only:bool=False,**kwargs:Any) -> 'CointRankResults':
    """ Returns the optimal cointigration rank for a multivariate process using the function from statsmodels: 
    https://www.statsmodels.org/dev/generated/statsmodels.tsa.vector_ar.vecm.select_coint_rank.html

    Args:
       mvf (MVForecaster): The MVForecaster object with series loaded to find the optimal rank for.
       train_only (bool): Default False. Whether to use the training data only in the test.
        **kwargs: Passed to the referenced statsmodels function.

    Returns:
        (CointRankResults): Object containing the cointegration rank suggested by the test and allowing a summary to be printed.

    >>> from scalecast.Forecaster import Forecaster
    >>> from scalecast.MVForecaster import MVForecaster
    >>> from scalecast.util import find_optimal_coint_rank
    >>> import pandas_datareader as pdr
    >>>
    >>> s1 = pdr.get_data_fred('UTUR',start='2000-01-01',end='2022-01-01')
    >>> s2 = pdr.get_data_fred('UNRATE',start='2000-01-01',end='2022-01-01')
    >>>
    >>> f1 = Forecaster(y=s1['UTUR'],current_dates=s1.index)
    >>> f2 = Forecaster(y=s2['UNRATE'],current_dates=s2.index)
    >>>
    >>> mvf = MVForecaster(f1,f2,names=['UTUR','UNRATE'])
    >>> coint_res = find_optimal_coint_rank(mvf,det_order=-1,k_ar_diff=8,train_only=True)
    >>> print(coint_res) # prints a report
    >>> rank = coint_res.rank # best rank
    """
    from statsmodels.tsa.vector_ar.vecm import select_coint_rank
    data = np.array([v.astype(float).values.copy() for k, v in mvf.y.items()]).T
    if train_only:
        data = data[:-mvf.test_length]
    
    return select_coint_rank(
        data,
        det_order=det_order,
        k_ar_diff=k_ar_diff,
        **kwargs,
    )

@_developer_utils.log_warnings
def find_statistical_transformation(
    f:'Forecaster',
    goal:list[Literal['stationary','seasonally_adj']]=['stationary'],
    train_only:bool=False,
    critical_pval:ConfInterval=.05,
    log:bool = True,
    m:int|Literal['auto']='auto',
    adf_kwargs:dict[str,Any] = {},
    **kwargs:Any,
) -> tuple['Transformer','Reverter']:
    """ Finds a set of transformations to achieve stationarity or seasonal adjustment, based on results from statistical tests.

    Args:
        f (Forecaster): The object that stores the series to test.
        goal (list-like): Default ['stationary']. One or multiple of 'stationary', 'seasonally_adj'. 
            Other options may be coming in the future.
            If more than one goal is passed, will try to satisfy all goals in the order passed.
            For stationary: uses an Augmented Dickey-Fuller test to determine if the series is stationary.
            If not stationary, returns a diff transformation and log transformation if log is True.
            For seasonall_adj: uses seasonal auto_arima to find the optimal seasonal diff.
        train_only (bool): Default False. Whether to use train set only in all statistical tests.
        log (bool): Default True. Whether to log and difference the series if it is found to be non-stationary or just difference.
            This will set itself to False if the lowest observed series value is 0 or lower.
        critical_pval (float): Default 0.05. The cutoff p-value to use to determine statistical signficance in the 
            Augmented Dickey-Fuller test and to run the auto_arima selection (substitutes for `alpha` arg).
        m (str or int): Default 'auto': The number of observations that counts one seasonal step.
            When 'auto', uses the M4 competition values: 
            for Hourly: 24, Monthly: 12, Quarterly: 4. everything else gets 1 (no seasonality assumed)
            so pass your own values for other frequencies.
        adf_kwargs (dict): Default {}. Keyword args to pass to the Augmented Dickey-Fuller test function.
            See the `maxlag`, `regression`, and `autolag` arguments from
            https://www.statsmodels.org/dev/generated/statsmodels.tsa.stattools.adfuller.html.
        **kwargs: Passed to the auto_arima() function when searching for optimal seasonal diff.

    Returns:
        (Transformer, Reverter): A `Transformer` object with the identified transforming functions and
        the `Reverter` object with the `Transformer` counterpart functions.

    >>> from scalecast.Forecaster import Forecaster
    >>> from scaleast.Pipeline import Pipeline, Transformer, Reverter
    >>> from scalecast.util import find_statistical_transformation
    >>> import pandas_datareader as pdr
    >>> 
    >>> def forecaster(f):
    >>>     f.add_covid19_regressor()
    >>>     f.auto_Xvar_select(cross_validate=True)
    >>>     f.set_estimator('mlr')
    >>>     f.manual_forecast()
    >>> df = pdr.get_data_fred(
    >>>     'HOUSTNSA',
    >>>     start='1959-01-01',
    >>>     end='2022-08-01'
    >>> )
    >>> f = Forecaster(
    >>>     y=df['HOUSTNSA'],
    >>>     current_dates=df.index,
    >>>     future_dates=24,
    >>>     test_length = .2,
    >>> )
    >>> f.set_validation_length(24)
    >>> transformer, reverter = find_statistical_transformation(
    >>>     f,
    >>>     goal=['stationary','seasonally_adj'],
    >>>     train_only=True,
    >>>     critical_pval = .01,
    >>> )
    >>> print(reverter) # see what transformers and reverters were chosen
    >>> pipeline = Pipeline(
    >>>   steps = [
    >>>       ('Transform',transformer),
    >>>       ('Forecast',forecaster),
    >>>       ('Revert',reverter),
    >>>   ],
    >>> )
    >>> f = pipeline.fit_predict(f)
    """
    from . import SeriesTransformer
    from . import Pipeline
    from .auxmodels import auto_arima
    def make_stationary(f,train_only,critical_pval,log,adf_kwargs,**kwargs):
        transformers = []
        stationary = f.adf_test(
            train_only=train_only,
            full_res=False,
            critical_pval=critical_pval,
            **adf_kwargs
        )
        if not stationary:
            if log and f.y.min() > 0:
                transformers += [('LogTransform',)]
            transformers += [('DiffTransform',1)]  
        return transformers

    def seasonally_adj(f,train_only,critical_pval,log,adf_kwargs,**kwargs):
        transformers = []
        if m == 1:
            return transformers

        auto_arima(f,m=m,seasonal=True,alpha=critical_pval,**kwargs)
        I = f.auto_arima_params['seasonal_order'][1]
        if I > 0:
            for i in range(1,I+1):
                transformers += [('DiffTransform',m)]

        return transformers

    f = copy.deepcopy(f)

    m = find_seasonal_length(m,f.freq)

    possible_args = {
        'stationary':make_stationary,
        'seasonally_adj':seasonally_adj,
    }
    bad_args = [g for g in goal if g not in possible_args]
    if len(bad_args) > 0:
        raise ValueError(f'Values passed to goal arg cannot be used: {bad_args}.')

    transformers = []
    reverters = []
    for g in goal:
        t = possible_args[g](
            f,
            train_only=train_only,
            critical_pval=critical_pval,
            log=log,
            adf_kwargs=adf_kwargs,
            **kwargs
        )
        transformers += t
    for t in transformers[::-1]:
        r = (t[0].replace('Transform','Revert'),)
        if len(t) > 1:
            r += t[1:]
        reverters.append(r)

    final_transformer = Pipeline.Transformer(transformers = transformers)
    final_reverter = Pipeline.Reverter(
        reverters = reverters, 
        base_transformer = final_transformer,
    )
    return final_transformer, final_reverter

@_developer_utils.log_warnings
def find_optimal_transformation(
    f:'Forecaster',
    estimator:AvailableModel=None,
    monitor:DefaultMetric='rmse',
    test_length:Optional[int]=None,
    train_length:Optional[int]=None,
    num_test_sets:int=1,
    space_between_sets:int = 1,
    lags:int|Literal['auto']='auto',
    try_order:list[TryTransformations] = ['detrend','seasonal_adj','boxcox','first_diff','first_seasonal_diff','scale'],
    boxcox_lambdas:list[float] = [-0.5,0,0.5],
    detrend_kwargs:dict[str,Any] = [{'loess':True},{'poly_order':1},{'poly_order':2}],
    scale_type:list[str] = ['Scale','MinMax','RobustScale'],
    m:NonNegativeInt|Literal['auto']='auto',
    model:Literal["additive", "add", "multiplicative", "mul"] = 'add',
    set_aside_test_set:bool=False,
    return_train_only:bool = False,
    verbose:True = False,
    **kwargs:Any,
) -> tuple['Transformer','Reverter']:
    """ Finds a set of transformations based on what maximizes forecast accuracy on some out-of-sample metric.
    Works by comparing each transformation individually and stacking the set of transformations that leads to the best
    performance. The estimator only uses series lags as inputs. When an attempted transformation fails, a warning is logged.
    The function uses `Pipeline.backtest()` to assure that the selected set of transformations is truly tested out-of-sample.

    Args:
        f (Forecaster): The Forecaster object that contains the series that will be transformed.
        estimator (str): One of `Forecaster.can_be_tuned`. The estimator to use to choose the best 
            transformations with. The default will read whatever is set to f.estimator.
        monitor (str or callable): Default 'rmse'. The error metric to minimize.
            If str, must exist in `util.metrics` 
            and accept only two arguments. If callable, must accept only two arguments 
            (an array of actuals and an array of forecasts) and return a float.
            If 'r2' is passed, this will monitor a negative r2 value.
        test_length (int): The amount of observations to hold out-of-sample. By default
            reads the number of dates in f.future_dates.
        train_length (int): The number of observations to train the model in each iteration.
            By default, uses all available observations that come before each test set.
        num_test_sets (int): Default 1. The number of test sets to iterate through. The final metric will be 
            an average across all test sets.
        space_between_sets (int): Default 1. The space between consecutive training sets.
            Not applicable when num_test_sets is 1.
        lags (str or int): Default 'auto'. The number of lags that will be used as inputs for the estimator.
            If 'auto', uses the value passed or assigned to m (one seasonal cycle). 
            If multiple values passed to m, uses the first.
        try_order (list-like): Default ['detrend','seasonal_adj','boxcox','first_diff','first_seasonal_diff','scale'].
            The transformations to try and also the order to try them in.
            Changing the order here can change the final transformations derived, since level will 
            be compared to the first transformation and if it is found to be better than level, it will
            carry over to be tried in conjunction with the next transformation and so forth.
            The default list contains all possible transformations for this function.
        boxcox_lambdas (list-like): Default [-0.5,0,0.5].
            The lambda values to try for a boxcox transformation.
            0 means natural log. Only up to one boxcox transformation will be selected.
        detrend_kwargs (list-like[dict]): Default 
            [{'loess':True},{'poly_order':1},{'poly_order':2}].
            The types of detrending to try. Only up to one one detrender will be selected.
        scale_type (list-like): Default ['Scale','MinMax','RobustScale']. The type of scaling to try.
            Only up to one scaler will be selected.
            Must exist a `SeriesTranformer.{scale_type}Transform()` function for this to work.
        m (int or str): Default 'auto'. The number of observations that counts one seasonal step.
            Ignored when seasonal_lags = 0.
            When 'auto', uses the M4 competition values:
            for Hourly: 24, Monthly: 12, Quarterly: 4. Everything else gets inferred if possible.
            If list, multiple adjustments will be tried and up to that many adjustments can be selected.
        model (str): Default 'add'. One of {"additive", "add", "multiplicative", "mul"}.
            The type of seasonal component. Only relevant for the 'seasonal_adj' option in try_order.
        set_aside_test_set (bool): Default False. 
            Whether to separate the test set specified in f.test_length during this process.
            Setting this to True prevents leakage when testing the forecasts out-of-sample.
        return_train_only (bool): Default False. Whether the returned selections should be set to train_only.
            All tries are completely out-of-sample but the returned transformations will not hold out the
            test-set in the Forecaster object when detrending, deseasoning, and scaling, so setting this to True
            can prevent leakage.
        verbose (bool): Default False. Whether to print info about the transformers/reverters being tried.
        **kwargs: Passed to the `Forecaster.manual_forecast()` function and possible values change based on which
            estimator is used.

    Returns:
        (Transformer, Reverter): A `Transformer` object with the identified transforming functions and
        the `Reverter` object with the `Transformer` counterpart functions.

    >>> from scalecast.Forecaster import Forecaster
    >>> from scaleast.Pipeline import Pipeline, Transformer, Reverter
    >>> from scalecast.util import find_optimal_transformation
    >>> import pandas_datareader as pdr
    >>> 
    >>> def forecaster(f):
    >>>     f.add_covid19_regressor()
    >>>     f.auto_Xvar_select(cross_validate=True)
    >>>     f.set_estimator('mlr')
    >>>     f.manual_forecast()
    >>> df = pdr.get_data_fred(
    >>>     'HOUSTNSA',
    >>>     start='1959-01-01',
    >>>     end='2022-08-01'
    >>> )
    >>> f = Forecaster(
    >>>     y=df['HOUSTNSA'],
    >>>     current_dates=df.index,
    >>>     future_dates=24,
    >>>     test_length = .2, # this will be monitored for performance
    >>> )
    >>> f.set_validation_length(24)
    >>> transformer, reverter = find_optimal_transformation(f)
    >>> print(reverter) # see what transformers and reverters were chosen
    >>> pipeline = Pipeline(
    >>>   steps = [
    >>>       ('Transform',transformer),
    >>>       ('Forecast',forecaster),
    >>>       ('Revert',reverter),
    >>>   ],
    >>> )
    >>> f = pipeline.fit_predict(f)
    """
    from . import Pipeline
    from ._Forecaster_parent import ForecastError, METRICS
    from .models import SKLearnUni

    def forecaster(f):
        f.add_ar_terms(lags)
        f.set_estimator(estimator)
        f.manual_forecast(**kwargs)

    def make_pipeline_fit_predict(f,transformer,reverter):
        tr = Pipeline.Transformer(transformers=transformer)
        re = Pipeline.Reverter(reverters=reverter,base_transformer=tr)
        pipeline = Pipeline.Pipeline(
            steps = [
                ('Transform',tr),
                ('Forecast',forecaster),
                ('Revert',re)
            ],
        )
        res = pipeline.backtest(
            f,
            n_iter = num_test_sets, 
            jump_back = space_between_sets,
            fcst_length = test_length,
            series_length = train_length,
            cis = False,
        )
        mets = backtest_metrics(res,mets=[monitor])
        if verbose:
            print(f'Last transformer tried:\n{tr.transformers}')
            print(f'Score ({monitor if isinstance(monitor,str) else monitor.__name__}): {mets.iloc[0,-1]}')
            print('-'*50)
        return mets.iloc[0,-1]

    estimator = f.estimator if estimator is None else estimator
    test_length = len(f.future_dates) if test_length is None else int(test_length)
    if test_length<=0:
        raise ValueError(
            'The argument test_length must be above 0 and is the number of future dates in the Forecaster object by default.'
            f' The value received is {test_length}. '
            "This can happen if you used the default test_length argument and didn't add future dates to the passed `Forecaster`."
        )

    f = copy.deepcopy(f)
    if set_aside_test_set:
        f.chop_from_front(f.test_length)
    
    f.set_metrics([monitor])
    f.drop_all_Xvars()
    f.history = {}

    m = find_seasonal_length(m,f.freq)
    lags = m if lags == 'auto' and not hasattr(m,'__len__') else m[0] if lags == 'auto' else lags

    if verbose:
        print(f'Using {estimator} model to find the best transformation set on {num_test_sets} test sets, each {test_length} in length.')
        if f.estimators.lookup_item(estimator).interpreted_model == SKLearnUni:
            print(f'All transformation tries will be evaluated with {lags} lags.')

    metric_store = METRICS.lookup_item(monitor)
    level_met = EvaluatedMetric(store=metric_store,score=make_pipeline_fit_predict(f,[],[]))
    final_transformer = []
    final_reverter = []

    exception_types = (IndexError,AttributeError,ValueError,ZeroDivisionError,ForecastError) # errors to pass over

    for tr in try_order:
        if tr == 'boxcox':
            for i, lmbda in enumerate(boxcox_lambdas):
                transformer = final_transformer[:]
                reverter = final_reverter[:]
                if i == 0:
                    met = level_met
                    best_transformer = transformer[:]
                    best_reverter = reverter[:]
                try:
                    transformer.append(('Transform',boxcox_tr,{'lmbda':lmbda}))
                    reverter.reverse(); reverter.append(('Revert',boxcox_re,{'lmbda':lmbda})); reverter.reverse()
                    comp_met = EvaluatedMetric(store=metric_store,score=make_pipeline_fit_predict(f,transformer,reverter))
                    if comp_met > met:
                        met = comp_met
                        best_transformer = transformer[:]
                        best_reverter = reverter[:]
                except exception_types as e:
                    warnings.warn(f'Lambda value of {lmbda} cannot be evaluated. error: {e}')
            final_transformer = best_transformer[:]
            final_reverter = best_reverter[:]
            level_met = met
        elif tr == 'detrend':
            for i, kw in enumerate(detrend_kwargs):
                transformer = final_transformer[:]
                reverter = final_reverter[:]
                if i == 0:
                    met = level_met
                    best_transformer = transformer[:]
                    best_reverter = reverter[:]
                try:
                    transformer.append(('DetrendTransform',kw))
                    reverter.reverse(); reverter.append(('DetrendRevert',)); reverter.reverse()
                    comp_met = EvaluatedMetric(store=metric_store,score=make_pipeline_fit_predict(f,transformer,reverter))
                    if comp_met > met:
                        met = comp_met
                        best_transformer = transformer[:]
                        best_reverter = reverter[:]
                except exception_types as e:
                    warnings.warn(f'Detrend_kwargs {kw} cannot be evaluated. error: {e}')
            final_transformer = best_transformer[:]
            final_reverter = best_reverter[:]
            level_met = met
        elif tr == 'first_diff':
            met = level_met
            transformer = final_transformer[:]
            reverter = final_reverter[:]
            best_transformer = transformer[:]
            best_reverter = reverter[:]
            try:
                transformer.append(('DiffTransform',1))
                reverter.reverse(); reverter.append(('DiffRevert',1,{'delete_attributes':False})); reverter.reverse()
                comp_met = EvaluatedMetric(store=metric_store,score=make_pipeline_fit_predict(f,transformer,reverter))
                if comp_met > met:
                    met = comp_met
                    best_transformer = transformer[:]
                    best_reverter = reverter[:]
            except exception_types as e:
                warnings.warn(f'Series first difference could not be evaluated. error: {e}')
            final_transformer = best_transformer[:]
            final_reverter = best_reverter[:]
            level_met = met
        elif tr in ('seasonal_adj', 'first_seasonal_diff'):
            if not hasattr(m,'__len__'):
                m = [m]
            for mi in m:
                if mi > 1:
                    met = level_met
                    transformer = final_transformer[:]
                    reverter = final_reverter[:]
                    best_transformer = transformer[:]
                    best_reverter = reverter[:]
                    try:
                        transformer += (
                            [('DiffTransform',mi)] if tr == 'first_seasonal_diff'
                            else [('DeseasonTransform',{'m':mi,'model':model})]
                        )
                        reverter.reverse()
                        reverter += (
                            [('DiffRevert',mi,{'delete_attributes':False})] if tr == 'first_seasonal_diff' 
                            else [('DeseasonRevert',{'m':mi})]
                        )
                        reverter.reverse()
                        comp_met = EvaluatedMetric(store=metric_store,score=make_pipeline_fit_predict(f,transformer,reverter))
                        if comp_met > met:
                            met = comp_met
                            best_transformer = transformer[:]
                            best_reverter = reverter[:]
                    except exception_types as e:
                        warnings.warn(f'Series seasonal adjustment could not be evaluated. error: {e}')
                    final_transformer = best_transformer[:]
                    final_reverter = best_reverter[:]
                    level_met = met
                else:
                    warnings.warn('Seasonal differences and adjustments cannot be evaluated when m = 1.')
        elif tr == 'scale':
            for i, s in enumerate(scale_type):
                transformer = final_transformer[:]
                reverter = final_reverter[:]
                if i == 0:
                    met = level_met
                    best_transformer = transformer[:]
                    best_reverter = reverter[:]
                try:
                    transformer.append((f'{s}Transform',))
                    reverter.reverse(); reverter.append((f'{s}Revert',)); reverter.reverse()
                    comp_met = EvaluatedMetric(store=metric_store,score=make_pipeline_fit_predict(f,transformer,reverter))
                    if comp_met > met:
                        met = comp_met
                        best_transformer = transformer[:]
                        best_reverter = reverter[:]
                except exception_types as e:                 
                    warnings.warn(f'{s} scaler cannot be evaluated. error: {e}')
            final_transformer = best_transformer[:]
            final_reverter = best_reverter[:]
            level_met = met
        else:
            warnings.warn(f'Value: {tr} found in the try_order list cannot be used and will be skipped.')
    
    final_transformer = Pipeline.Transformer(transformers = final_transformer)
    final_reverter = Pipeline.Reverter(reverters = final_reverter,base_transformer = final_transformer)

    if return_train_only:
        has_train_only = (
            'DetrendTransform',
            'DeseasonTransform',
            'ScaleTransform',
            'MinMaxTransform',
            'RobustScaleTransform',
        )
        for i, t in enumerate(final_transformer.transformers):
            if t[0] in has_train_only:
                if len(t) == 1 or not isinstance(t[-1],dict):
                    final_transformer.transformers[i] = t + ({'train_only':True},)
                else:
                    t[-1]['train_only'] = True

    if verbose:
        print(f'Final Selection:\n{final_transformer.transformers}')
    
    return final_transformer, final_reverter 

def Forecaster_with_missing_vals(
    y:Sequence[float,int],
    current_dates:list[DatetimeLike],
    desired_frequency:Optional[str] = None,
    fill_strategy:float|Literal['linear_interp', 'moving_average', 'moving_seasonal_average', 'impute_pool','ffill','bfill'] = 0.0,
    impute_value_pool:Optional[list[float]] = None,
    m:Optional[int] = None,
    impute_lookback:Optional[int] = None,
    add_noise:bool = False,
    noise_value_pool:Optional[list[float]] = None,
    noise_std:Optional[float] = None,
    noise_lookback:Optional[int] = None,
    cannot_be_below:Optional[float] = None,
    cannot_be_above:Optional[float] = None,
    first_ob_strategy:Literal['drop','ignore']|FFillOption = 'drop',
    random_seed:Optional[int] = None,
    **kwargs:Any,
) -> 'Forecaster':
    """ Imputes missing values in a given time series such that the result has a user-specified 
    date frequency and/or no remaining null values. If you pass no missing values through this function,
    it will not raise errors.

    Args:
        y (collection): An array of all observed values. Can include NAs for dates in which the values
            are unknown.
        current_dates (collection): An array of all observed dates. 
            Must be same length as y and in the same sequence. 
        desired_frequency (str): The desired frequency of the resulting Forecaster object.
            If this is left unspecified and a frequency cannot be inferred, the resulting
            object will not have a logical frequency. See available values here:
            https://pandas.pydata.org/docs/user_guide/timeseries.html#timeseries-offset-aliases
        fill_strategy (float or str): Default 0.0. 
            If str, must be one of 
            {'linear_interp', 'moving_average', 'moving_seasonal_average', 'impute_pool'}.
            If not one of those values, will be passed to the `df.fillna()` method from pandas
            (valid values include 'ffill' and 'bfill'). Therefore, the default fills with 0.
        m (int): Optional. The number of steps that count one seasonal cycle if using a seasonal fill strategy.
            If left unspecified, will attempt to be inferred. If it cannot be inferred, will raise an error.
        impute_value_pool (collection): Optional. The pool of values to use when `fill_strategy = 'impute_pool'`.
        impute_lookback (int): Required when `fill_strategy in ('moving_average','moving_seasonal_average')`.
            The lookback to use when imputing a moving average to missing values. If using 'moving_seaosnal_average', 
            make sure to include at least one full seasonal cycle in the lookback. Must be 1 or greater. If there are not enough
            observations to create a seasonal fill, will use all available observations for a normal moving average and raise a warning.
        add_noise (bool): Default False. Whether to add random noise to the imputed values.
        noise_value_pool (collection): Optional. The pool of values to randomly choose from when adding noise.
            The noise will add the imputed value with a random draw from this pool to come up with the final value.
            Specifying this argument overrides any of the subsequent noise-related arguments 
            (noise_std, noise_lookback, etc.).
        noise_std (float): Optional. The standard deviation to use when adding a noise to the values.
            Assumes a normal distribution where the mean is the value imputed.
        noise_lookback (int): Optional. Must be 2 or greater.
            If adding noise, the lookback period before the missing obs
            to use to add the noise, assuming a normal distribution with the standard deviation from the lookback.
            If this is larger than the number of observations before a given missing observation, will use all
            observations before the missing one. If this and all the other noise-related arguments
            are left unspeficied, uses all observations before each missing one to find the 
            standard deviation. If the first observation(s) is missing, no noise is given to it.
        cannot_be_below (float): Optional. A minimum value that the final imputation cannot drop below.
        cannot_be_above (float): Optional. A maximum value that the final imputation cannot be above.
        first_ob_strategy (str): Default 'drop'. What to do if the first observation(s) is null. Default will drop.
            Other options include 'ignore', which could cause unexpected results depending on the employed strategy.
            Can also start with 'fill_', where the next digits will be used to create a static fill 
            ('fill_0' fills with 0, for example).
        random_seed (int): Optional. A random seed to set for reproducible results.
        **kwargs: Passed to the Forecaster object (https://scalecast.readthedocs.io/en/latest/Forecaster/Forecaster.html#src.scalecast.Forecaster.Forecaster.__init__)

    Returns:
         (Forecaster): A Forecaster object with missing dates/values filled in.

    >>> # using the function with null values in y
    >>> f = Forecaster_with_missing_vals(
    >>>    y = [1,2,np.nan,4],
    >>>    current_dates=['2020-01-01','2020-01-02','2020-01-03','2020-01-04'],
    >>>    fill_strategy = 'linear_interp',
    >>> ) # replaces missing val with 3
    >>> # using the function with missing dates
    >>> f = Forecaster_with_missing_vals(
    >>>    y = [1,2,4],
    >>>    current_dates=['2020-01-01','2020-01-02','2020-01-04'], # missing '2020-01-03'
    >>>    desired_frequency = 'D', # tell it to use daily frequency
    >>>    fill_strategy = 'linear_interp',
    >>> ) # adds 3 to the 2nd index position in y and adds '2020-01-03' to 2nd index position in current_dates
    """
    from .Forecaster import Forecaster
    if random_seed is not None:
        random.seed(random_seed)

    valid_strategies = [
        'linear_interp',
        'moving_average', 
        'moving_seasonal_average', 
        'impute_pool',
    ]

    ts_df = pd.DataFrame({
        'Date':pd.to_datetime(current_dates),
        'y':[i for i in y],
    })

    if desired_frequency is not None:
        full_ts_df = pd.DataFrame({
            'Date':pd.date_range(
                start=ts_df['Date'].min(),
                end=ts_df['Date'].max(),
                freq = desired_frequency,
            )
        })
        ts_df = full_ts_df.merge(
            ts_df, on = 'Date', how = 'left',
        )
    ts_df['missing'] = ts_df['y'].isnull().astype(int)
    if ts_df['missing'].sum() > 0:
        if (
            fill_strategy not in valid_strategies 
            and not add_noise 
            and (
                not np.isnan(ts_df['y'].values[0])
                or first_ob_strategy == 'ignore'
            )
        ):
            try:
                ts_df.fillna(float(fill_strategy),inplace=True)
            except ValueError:
                ts_df = getattr(ts_df,fill_strategy)
        else:
            ts_df['missing_number'] = (
                (ts_df['missing'] == 1) 
                & (
                    (ts_df['missing'].shift() == 0)
                    | (ts_df['missing'].shift().isnull())
                )
            ).astype(int).cumsum()
            ts_df['missing_number'] = ts_df[['missing','missing_number']].apply(
                lambda x: x.iloc[1] if x.iloc[0] == 1 else 0,
                axis = 1,
            )

            if np.isnan(ts_df['y'].values[0]):
                if first_ob_strategy == 'ignore':
                    if fill_strategy in [valid_strategies[0],valid_strategies[2]]: # li, sma
                        raise ValueError(
                            'Cannot perform `fill_strategy="linear_interp" '
                            'or `fill_strategy="moving_seasonal_average"` when the first observation is missing.'
                        )
                else:
                    if first_ob_strategy == 'drop':
                        ts_df = ts_df.loc[ts_df['missing_number'] != 1].reset_index(drop=True)
                    elif first_ob_strategy.startswith('fill_'):
                        ts_df.loc[ts_df['missing_number'] == 1,'y'] = float(first_ob_strategy.split('fill_')[-1])
                    else:
                        raise ValueError(f'Unexpected value passed to argument first_ob_strategy: {first_ob_strategy}.')
                    ts_df['missing_number'] = ts_df['missing_number'].apply(lambda x: max(0,x-1))

            if fill_strategy in [valid_strategies[0],valid_strategies[2]]: # li, sma
                ts_df['m'] = 0
                ts_df['b'] = 0
                ts_df['x'] = 0

                dfs = []
                for cn in ts_df['missing_number'].unique():
                    if cn == 0:
                        dfs.append(ts_df.loc[ts_df['missing_number'] == cn])
                    else:
                        ts_df_tmp = ts_df.loc[ts_df['missing_number'] == cn].copy()
                        ts_df_tmp['x'] = 1
                        ts_df_tmp['x'] = ts_df_tmp['x'].cumsum()
                        y1 = ts_df.loc[ts_df_tmp.index[0] - 1,'y']
                        y2 = ts_df.loc[ts_df_tmp.index[-1] + 1,'y']
                        ts_df_tmp['m'] = (y2-y1) / (ts_df_tmp.shape[0] + 1)
                        ts_df_tmp['b'] = y1
                        dfs.append(ts_df_tmp)
                ts_df = pd.concat(dfs).sort_values('Date').reset_index(drop=True)

            for i, v in ts_df.iterrows():
                if np.isnan(v['y']):
                    if fill_strategy == valid_strategies[0]: # li
                        ts_df.loc[i,'y'] = v['m'] * v['x'] + v['b'] # y = mx + b :)
                    
                    elif fill_strategy in (valid_strategies[1],valid_strategies[2]): # ma, sma
                        if v['missing_number'] > 0:
                            _developer_utils.descriptive_assert(
                                impute_lookback is None or impute_lookback >= 1, 
                                ValueError, 
                                f'impute_lookback must be 1 or greater, got {impute_lookback}.'
                            )
                        impute_lookback_i = impute_lookback
                        fill_strategy_i = fill_strategy
                        ma_pool = ts_df.loc[
                            (ts_df['Date'] < v['Date']) 
                            & (ts_df['missing_number'] < v['missing_number'])
                        ]

                        if fill_strategy == valid_strategies[2]: # sma
                            if v['missing_number'] == 1:
                                if m is None:
                                    m = find_seasonal_length('auto',pd.infer_freq(ts_df['Date']))
                            if impute_lookback_i is None:
                                impute_lookback_i = ma_pool.shape[0]
                            else:
                                ma_pool = ma_pool.iloc[-int(impute_lookback_i):]
                            if m > impute_lookback_i:
                                warnings.warn(
                                    'Not enough observations to impute missing value with '
                                    'a seaonal moving average for date {}. '
                                    'Defaulting to a normal moving average.'.format(v['Date']),
                                    category = Warning,
                                )
                                fill_strategy_i = valid_strategies[1]
                        else:
                            if impute_lookback is not None:
                                ma_pool = ma_pool.iloc[-int(impute_lookback):]

                        if fill_strategy_i == valid_strategies[1]: # ma
                            ts_df.loc[i,'y'] = ma_pool['y'].mean()
                        else: # sma
                            ts_df.loc[i,'y'] = np.mean(
                                [j for i, j in enumerate(ma_pool['y'].values[::-1]) if (i + int(v['x'])) % m == 0]
                            )
                    
                    elif fill_strategy == valid_strategies[3]: # pool
                        ts_df.loc[i,'y'] = np.random.choice(impute_value_pool)
                    else:
                        try:
                            ts_df.loc[i,'y'] = float(fill_strategy)
                        except ValueError:
                            ts_df_cut = ts_df.loc[ts_df['Date'] <= v['Date']].copy()
                            ts_df_cut['y'] = ts_df_cut['y'].fillna(method = fill_strategy)
                            ts_df.loc[i,'y'] = ts_df_cut['y'].values[-1]
                    
                    if add_noise:
                        if noise_value_pool is not None:
                            ts_df.loc[i,'y'] += np.random.choice(noise_value_pool)
                        else:
                            if noise_std is None:
                                noise_std_df = ts_df.loc[
                                    (ts_df['Date'] < v['Date']) 
                                    & (ts_df['missing_number'] < v['missing_number'])
                                ]
                                if noise_lookback is not None:
                                    _developer_utils.descriptive_assert(
                                        noise_lookback >= 2,
                                        ValueError,
                                        f'noise_lookback must be 2 or greater, got {noise_lookback}.'
                                    )
                                    noise_std_df = noise_std_df.iloc[-int(noise_lookback):]
                                noise_std = noise_std_df['y'].std()
                            
                            ts_df.loc[i,'y'] = np.random.normal(ts_df.loc[i,'y'], noise_std)
                    if cannot_be_below is not None:
                        ts_df.loc[i,'y'] = max(ts_df.loc[i,'y'], cannot_be_below)
                    if cannot_be_above is not None:
                        ts_df.loc[i,'y'] = min(ts_df.loc[i,'y'], cannot_be_above)

    return Forecaster(
        y = ts_df['y'].values,
        current_dates = ts_df['Date'].values,
        **kwargs,
    )

def backtest_for_resid_matrix(
    *fs:'Forecaster',
    pipeline:'Pipeline',
    alpha:ConfInterval = 0.05,
    bt_n_iter:Optional[PositiveInt] = None,
    jump_back:PositiveInt = 1,
    **kwargs:Any,
) -> list[dict[str,pd.DataFrame]]:
    """ Performs a backtest on one or more Forecaster objects using pipelines.
    Specifically, performs a backtest so that a residual matrix to make dynamic intervals
    can easily be obtained. (See `util.get_backtest_resid_matrix()` and 
    `util.overwrite_forecast_intervals()`).

    Args:
        *fs (Forecaster): The objects that contain the evaluated forecasts.
            Send one if univariate forecasting with the `Pipeline` class, 
            more than one if multivariate forecasting with the `MVPipeline` class.
        pipeline (Pipeline or MVPipeline): The pipeline to send `*fs` through.
        alpha (float): Default 0.05. The level that confidence intervals need to be evaluated at. 0.05 = 95%.
        bt_n_iter (int): Optional. The number of iterations to backtest. If left unspecified, chooses 1/alpha, 
            the minimum needed to set reliable conformal intervals.
        jump_back (int): Default 1. The space between consecutive training sets in the backtest.
        **kwargs: Passed to `Pipeline.backtest()`.


    Returns:
        (List[Dict[str,pd.DataFrame]]): The results from each model and backtest iteration.
        Each dict element of the resulting list corresponds to the `Forecaster` objects in the order
        they were passed (will be length 1 if univariate forecasting). Each key of each dict is either 'Actuals', 'Obs',
        or the name of a model that got backtested. Each value is a DataFrame with the iteration values.
        The 'Actuals' frame has the date information and are the actuals over each forecast horizon. 
        The 'Obs' frame has the actual historical observations to make each forecast, 
        back padded with NA values to make each array the same length.
    """
    bt_n_iter = int(round(1/alpha)) if bt_n_iter is None else bt_n_iter
    if bt_n_iter < round(1/alpha):
        raise ValueError(
            'bt_n_iter must be at least 1/alpha.'
            f' alpha is {alpha} and bt_n_iter is {bt_n_iter}.'
            f' bt_n_iter must be at least {1/alpha:.0f} to successfully '
            f' backtest {1-alpha:.0%} confidence intervals.'
        )
    bt_results = pipeline.backtest(
        *fs,
        n_iter=bt_n_iter,
        jump_back = jump_back,
        test_length = 0,
        cis = False,
        **kwargs,
    )
    return bt_results

def get_backtest_resid_matrix(backtest_results:list[dict[str,pd.DataFrame]]) -> list[dict[str,np.ndarray]]:
    """ Converts results from a backtest pipeline into a matrix of residuals.
    Each row in this residual is for a backtest iteration and the columns are a forecast step.

    Args:
        backtest_results (list): The output returned from `Pipeline.backtest()` or `MVPipeline.backtest()`.
            Recommend to obtain this from running `util.backtest_for_resid_matrix()` and to pass
            the results to `util.overwrite_forecast_intervals()`.

    Returns:
        (list[dict[str,numpy.ndarray]]): A list where each element corresponds to the given Forecaster object
        in a backtest. The elements of the list are dictionaries where each key is an evaluated model name and
        each value is a numpy matrix of the appropriate dimensions that can be used to determine a dynamic prediction interval.
    """
    mats = []
    for btr in backtest_results:
        mat = {
            m:np.zeros((btr['Obs'].shape[1],btr['Actuals'].shape[0]))
            for m in btr if m not in ('Actuals','Obs')
        }
        for m, v in btr.items():
            if m in ('Actuals','Obs'):
                continue
            for i in range(mat[m].shape[0]):
                mat[m][i,:] = btr['Actuals'][f'Iter{i}Vals'].values - v[f'Iter{i}Fcst'].values
        mats.append(mat)
    return mats

def overwrite_forecast_intervals(
    *fs:Forecaster,
    backtest_resid_matrix:list[dict[str,np.ndarray]],
    models:Optional[list[str]]=None,
    alpha:ConfInterval = .05
) -> Unpack[Forecaster]:
    """ Overwrites naive forecast intervals stored in passed Forecaster objects with dynamic intervals.
    Overwrites future predictions only; does not overwrite intervals for test-set prediction intervals.

    Args:
        *fs (Forecaster): The objects that contain the evaluated forecasts to overwrite confidence intervals.
        backtest_resid_matrix (list): The output returned from `util.get_backtest_resid_matrix()`.
        models (list): Optional. The models to overwrite intervals for. By default, overwrites all
            models found in backtest_resid_matrix.
        alpha (float): Default 0.05. The level that confidence intervals need to be evaluated at. 0.05 = 95%.
            Use the same or larger value passed to backtest_for_resid_matrix() or else this will fail.
    """
    for f, matrix in zip(fs,backtest_resid_matrix):
        i = 0
        models = matrix.keys() if models is None else models
        for m, mat in matrix.items():
            if m not in models:
                continue
            elif i == 0 and mat.shape[0] < round(1/alpha):
                raise ValueError(
                    f'Not enough backtested observations to evaluate confidence intervals at the {1-alpha:.0%} level.'
                    ' Please set alpha to whatever it was set to when running backtest_for_resid_matrix().'
                )
                i += 1
            percentiles = np.percentile(np.abs(mat),100*(1-alpha),axis=0)
            f.history[m]['UpperCI'] = np.array(f.history[m]['Forecast']) + percentiles
            f.history[m]['LowerCI'] = np.array(f.history[m]['Forecast']) - percentiles
            f.history[m]['CILevel'] = 1-alpha
    return fs

def gen_rnn_grid(
    layer_tries:int=5,
    min_layer_size:int=1,
    max_layer_size:int=3,
    layer_cell_pool:list[str]=['LSTM'],
    units_pool:list[int]=[8,16,32,64],
    activation_pool:list[str]=['relu','tanh'],
    dropout_pool:list[float]=[0.0],
    uniform_layer_cells:bool=True,
    uniform_units:bool=True,
    uniform_activations:bool=True,
    uniform_dropout:bool=True,
    verbose:bool|int=0,
    random_seed:int = None,
    **kwargs:Any,
) -> dict[str,Any]:
    """ Randomly generates an RNN grid for tuning hyperparameters.
    The resulting grid may be very large, so it is generally a good idea to use `Forecaster.limit_grid_size`
    to make it feasible to use with tuning/cross validation.

    Args:
        layer_tries (int): Default 5. How many layer tries to place into the grid.
        min_layer_size (int): Default 1. The smallest possible hidden layer structure will be this size.
        max_layer_size (int): Default 3. The largest possible hidden layer structure will be this size.
        layer_cell_pool (collection['LSTM'|'SimpleRNN']): Default ['LSTM']. The possible cell types to try.
        units_pool (collection[int]): Default [8,16,32,64]: The possible unit size for each cell.
        activation_pool (collection[str|tf.Activation]): Default ['relu','tanh']. The possible activation functions to try.
        dropout_pool (collection): Default [0]. The possible dropout values to add to each layer.
        uniform_layer_cells (bool): Default True. Whether each cell should be the same in all layers.
        uniform_units (bool): Default True. Whether each layer should have the same unit sizes.
        uniform_activations (bool): Default True. Whether each layer should have the same activation function.
        uniform_dropout (bool): Default True. Whether each layer should have the same dropout value.
        verbose (int): Default 0. The verbosity of the resulting models. 1 for fully verbose, 2 for medium verbose, 0 for no verbosity.
        random_seed (int): Optional. Set a random seed for consistent results.
        **kwargs: Other keyword arguments to add to each hyperparameter. 
            Can include callbacks for early stopping, optimizer, lags, etc. 
            If wanting to try multiple values for a given keyword, pass a list/collection type.
            See: https://scalecast.readthedocs.io/en/latest/Forecaster/_forecast.html#rnn

    Returns:
        (Dict): The resulting hyperparamter grid.

    >>> def forecaster(f):
    >>>     f.set_estimator('rnn')
    >>>     f.add_ar_terms(12)
    >>>     f.add_seasonal_regressors('month',raw=False,sincos=True)
    >>>     for i in range(10):
    >>>         grid = gen_rnn_grid()
    >>>         f.ingest_grid(grid)
    >>>         f.limit_grid_size(10)
    >>>         f.cross_validate(test_length=24,k=3)
    >>>         f.auto_forecast(call_me=f'lstm_{i}')
    """
    if random_seed is not None:
        np.random.seed(random_seed)
    hps = {}
    hps['verbose'] = [verbose]
    for k,v in kwargs.items():
        hps[k] = [v] if not hasattr(v,'__len__') else v

    pool_choices = {
        'uniform_layer_cells':'layer_cell_pool',
        'uniform_units':'units_pool',
        'uniform_activations':'activation_pool',
        'uniform_dropout':'dropout_pool',
    }

    hps['layers_struct'] = []
    for i in range(layer_tries):
        layer = []
        pool_values = {}
        for k, v in pool_choices.items():
            pool_values[v] = locals()[v]
            if locals()[k] and len(locals()[v]) > 1:
                pool_values[v] = [np.random.choice(locals()[v])]

        for _ in range(np.random.choice(np.arange(min_layer_size,max_layer_size+1))):
            layer.append((
                str(np.random.choice(pool_values['layer_cell_pool'])),
                {
                    'units':int(np.random.choice(pool_values['units_pool'])),
                    'activation':str(np.random.choice(pool_values['activation_pool'])),
                    'dropout':float(np.random.choice(pool_values['dropout_pool'])),
                },
            ))

        hps['layers_struct'].append(layer)

    return hps


from __future__ import annotations
from functools import wraps
import logging
import warnings
import numpy as np
from scipy import stats
from typing import Literal, Any, Sequence, TYPE_CHECKING
from .types import ConfInterval, DynamicTesting
if TYPE_CHECKING:
    from ._Forecaster_parent import Forecaster_parent

class _developer_utils:
    @staticmethod
    def log_warnings(func:callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with warnings.catch_warnings(record=True) as warn_list:
                warnings.simplefilter("always")
                result = func(*args, **kwargs)
                for warn in warn_list:
                    logging.warning(warn.message)
            return result
        return wrapper
    
    @staticmethod
    def descriptive_assert(statement:bool, ErrorType:Exception, error_message:str):
        try:
            assert statement
        except AssertionError:
            raise ErrorType(error_message)

    @staticmethod      
    def _return_na_if_len_zero(y:Sequence[float],pred:Sequence[float],func:callable):
        if len(pred) == 0:
            return np.nan
        else:
            return func(y,pred[-len(y):])

    @staticmethod
    def _set_ci_step(f:'Forecaster_parent',s:int):
        return stats.norm.ppf(1 - (1 - f.cilevel) / 2) * s

    @staticmethod
    def _check_train_only_arg(f:'Forecaster_parent', train_only:bool):
        _developer_utils.descriptive_assert(isinstance(train_only, bool), ValueError, f"train_only must be True or False, got {train_only} of type {type(train_only)}.")
        _developer_utils.descriptive_assert(not train_only or f.test_length > 0, ValueError, "train_only cannot be True when test_length is 0.")

    @staticmethod
    def _warn_about_not_finding_cis(m:str):
        warnings.warn(
            f'Confidence intervals not found for {m}. '
            'To turn on confidence intervals for future evaluated models, call the eval_cis() method.'
            ,category=Warning
        )

    @staticmethod
    def _reshape_func_input(x:Sequence[float],func:callable):
        x = np.array(x).reshape(-1,1)
        if x.shape[0] == 0:
            return []
        else:
            return func(x)[:,0]

    @staticmethod
    def _select_reg_for_direct_forecasting(f:'Forecaster_parent'):
        return {
            k:v.to_list() 
            for k, v in f.current_xreg.items() 
            if (
                np.isnan(f.future_xreg[k]).sum() == 0 
                and len(f.future_xreg[k]) == len(f.future_dates)
            )
        }

class NamedBoxCox:
    def __init__(self,name:str,transform:bool):
        self.name = name
        self.transform = transform

    def __call__(self,x,lmbda):
        if self.transform:
            return [(i**lmbda - 1) / lmbda for i in x] if lmbda != 0 else [np.log(i) for i in x]
        else:
            return [(i*lmbda + 1)**(1/lmbda) for i in x] if lmbda != 0 else [np.exp(i) for i in x]

    def __repr__(self):
        return self.name
    
def _tune_test_forecast(
    f:"Forecaster_parent",
    models:list[str],
    cross_validate:bool,
    dynamic_tuning:DynamicTesting,
    dynamic_testing:DynamicTesting,
    limit_grid_size:int|ConfInterval,
    suffix:str,
    error:Literal['raise','warn','ignore'],
    min_grid_size:int = 1,
    feature_importance:bool = False,
    fi_try_order:list[str] = None,
    use_progress_bar:bool = False,
    **cvkwargs:Any,
):
    if use_progress_bar: # notebooks only get progress bar
        from tqdm.notebook import tqdm
    else:
        tqdm = list

    for m in tqdm(models):
        call_me = m if suffix is None else m + suffix
        f.set_estimator(m)
        if limit_grid_size is not None:
            f.ingest_grid(m)
            f.limit_grid_size(n=limit_grid_size,min_grid_size=min_grid_size)
        if cross_validate:
            f.cross_validate(dynamic_tuning=dynamic_tuning, **cvkwargs)
        else:
            f.tune(dynamic_tuning=dynamic_tuning)
        try:
            f.auto_forecast(
                dynamic_testing=dynamic_testing,
                call_me=call_me,
            )
        except Exception as e:
            if error == 'raise':
                raise
            elif error == 'warn':
                warnings.warn(
                    f"{m} model could not be evaluated. "
                    f"Here's the error: {e}",
                    category=Warning,
                )
                continue
            elif error == 'ignore':
                continue
            else:
                raise ValueError(f'Value passed to error arg not recognized: {error}')
        if feature_importance:
            if fi_try_order is None:
                f.save_feature_importance()
            else:
                f.save_feature_importance(try_order=fi_try_order)

boxcox_tr = NamedBoxCox(name='BoxcoxTransform',transform=True)
boxcox_re = NamedBoxCox(name='BoxcoxRevert',transform=False)